# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""The `DataQualityFlag` and `DataQualityDict`.

The `DataQualityFlag` represents an annotated set of data-quality segments
indicating something about the state of a laser-interferometer
gravitational-wave detector in a given time interval.

The `DataQualityDict` is just a `dict` of flags, provided as a convenience
for handling multiple flags over the same global time interval.
"""

from __future__ import annotations

import contextlib
import json
import operator
import os
import re
import textwrap
import warnings
from concurrent.futures import ThreadPoolExecutor
from copy import (
    copy as shallowcopy,
    deepcopy,
)
from functools import (
    reduce,
    wraps,
)
from io import BytesIO
from math import (
    ceil,
    floor,
)
from typing import TYPE_CHECKING
from urllib.parse import urlparse

from astropy.table import Row as AstropyTableRow
from dqsegdb2.query import query_segments
from dqsegdb2.utils import get_default_host
from gwosc import timeline
from numpy import inf
from requests import codes as http_codes
from requests.exceptions import (
    HTTPError,
    RequestException,
)

from ..io.registry import UnifiedReadWriteMethod
from ..time import (
    LIGOTimeGPS,
    to_gps,
)
from ..utils.misc import if_not_none
from .connect import (
    DataQualityDictRead,
    DataQualityDictWrite,
    DataQualityFlagRead,
    DataQualityFlagWrite,
)
from .segments import (
    Segment,
    SegmentList,
)

if TYPE_CHECKING:
    from collections.abc import (
        Callable,
        Iterable,
    )
    from typing import (
        ParamSpec,
        Self,
        SupportsFloat,
        TypeVar,
    )

    import astropy.table
    import igwn_ligolw

    from ...plot import Plot
    from ...time import SupportsToGps

    P = ParamSpec("P")
    R = TypeVar("R")

    SegmentListLike = Iterable[tuple[SupportsFloat, SupportsFloat]]

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__all__ = [
    "DataQualityDict",
    "DataQualityFlag",
]

DEFAULT_SEGMENT_SERVER = get_default_host()

IFO_TAG_VERSION_REGEX = re.compile(
    r"\A(?P<ifo>[A-Z]\d):(?P<tag>[^/]+):(?P<version>\d+)\Z")
IFO_TAG_REGEX = re.compile(r"\A(?P<ifo>[A-Z]\d):(?P<tag>[^/]+)\Z")
TAG_VERSION_REGEX = re.compile(r"\A(?P<tag>[^/]+):(?P<version>\d+)\Z")


# -- utilities -----------------------

def _parse_query_segments(
    args: tuple[
        SegmentList | tuple[SupportsToGps, SupportsToGps]
    ] | tuple[SupportsToGps, SupportsToGps],
    func: Callable,
) -> SegmentList:
    """Parse ``args`` for query_dqsegdb() or query_segdb().

    Returns a SegmentList in all cases
    """
    # user passed SegmentList
    if len(args) == 1 and isinstance(args[0], SegmentList):
        return args[0]

    # otherwise unpack two arguments as a segment
    if len(args) == 1:
        args = args[0]

    # if not two arguments, panic
    try:
        start, end = args
    except ValueError as exc:
        exc.args = (f"{func.__name__}() takes 2 arguments for start and end GPS time, "
                    "or 1 argument containing a Segment or SegmentList",)
        raise

    # return list with one Segment
    return SegmentList([Segment(to_gps(start), to_gps(end))])


def _check_on_error(func: Callable[P, R]) -> Callable[P, R]:
    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        on_error = kwargs.get("on_error", "raise")
        # check on_error flag
        if on_error not in {"raise", "warn", "ignore"}:
            msg = (
                "on_error must be one of 'raise', 'warn', or 'ignore', "
                f"not '{on_error}'"
            )
            raise ValueError(msg)
        return func(*args, **kwargs)

    return wrapper



# -- DataQualityFlag -----------------

class DataQualityFlag:
    """A representation of a named set of segments.

    Parameters
    ----------
    name : str, optional
        The name of this flag.
        This should be of the form {ifo}:{tag}:{version}, e.g.
        'H1:DMT-SCIENCE:1'. Use `label` for human-readable names.

    active : `SegmentList`, optional
        A list of active segments for this flag

    known : `SegmentList`, optional
        A list of known segments for this flag

    label : `str`, optional
        Human-readable name for this flag, e.g. ``'Science-mode'``

    category : `int`, optional
        Veto category for this flag.

    description : `str`, optional
        Human-readable description of why this flag was created.

    isgood : `bool`, optional
        Do active segments mean the IFO was in a good state?
    """

    _EntryClass: type[Segment] = Segment
    _ListClass: type[SegmentList] = SegmentList

    def __init__(
        self,
        name: str | None = None,
        active: SegmentListLike | None = None,
        known: SegmentListLike | None = None,
        *,
        label: str | None = None,
        category: int | None = None,
        description: str | None = None,
        isgood: bool = True,
        padding: tuple[float | None, float | None] | None = None,
    ) -> None:
        """Define a new DataQualityFlag."""
        self.name = name
        self.known = known
        self.active = active
        self.label = label
        self.category = category
        self.description = description
        self.isgood = isgood
        self.padding = padding

    # -- utilities -------------------

    def _to_segmentlist(
        self,
        segmentlist: SegmentListLike | None,
    ) -> SegmentList:
        """Convert input to SegmentList of the correct type."""
        if segmentlist is None:
            return self._ListClass()
        return self._ListClass(self._EntryClass(*seg) for seg in segmentlist)

    # -- properties ------------------

    @property
    def name(self) -> str | None:
        """The name associated with this flag.

        This normally takes the form {ifo}:{tag}:{version}. If found,
        each component is stored separately the associated attributes.

        :type: `str`
        """
        return self._name

    @name.setter
    def name(self, name: str | None) -> None:
        self._name = name
        try:
            self._parse_name(name)
        except ValueError:
            self._parse_name(None)

    @property
    def ifo(self) -> str | None:
        """The interferometer associated with this flag.

        This should be a single uppercase letter and a single number,
        e.g. ``'H1'``.
        """
        return self._ifo

    @ifo.setter
    def ifo(self, ifoname: str | None) -> None:
        self._ifo = ifoname

    @property
    def tag(self) -> str | None:
        """The tag (name) associated with this flag.

        This should take the form ``'AAA-BBB_CCC_DDD'``, i.e. where
        each component is an uppercase acronym of alphanumeric
        characters only, e.g. ``'DCH-IMC_BAD_CALIBRATION'`` or
        ``'DMT-SCIENCE'``.
        """
        return self._tag

    @tag.setter
    def tag(self, tag: str | None) -> None:
        self._tag = tag

    @property
    def version(self) -> int | None:
        """The version number of this flag.

        Each flag in the segment database is stored with a version
        integer, with each successive version representing a more
        accurate dataset for its known segments than any previous.
        """
        return self._version

    @version.setter
    def version(self, v: int | None) -> None:
        self._version = if_not_none(int, v)

    @property
    def label(self) -> str | None:
        """A human-readable label for this flag.

        For example: ``'Science-mode'``.
        """
        return self._label

    @label.setter
    def label(self, lab: str | None) -> None:
        self._label = lab

    @property
    def active(self) -> SegmentList:
        """The set of segments during which this flag was active."""
        return self._active

    @active.setter
    def active(self, segmentlist: SegmentListLike | None) -> None:
        self._active = self._to_segmentlist(segmentlist)

    @active.deleter
    def active(self) -> None:
        self._active = self._ListClass()

    @property
    def known(self) -> SegmentList:
        """The segments during which this flag was known."""
        return self._known

    @known.setter
    def known(self, segmentlist: SegmentListLike | None) -> None:
        self._known = self._to_segmentlist(segmentlist)

    @known.deleter
    def known(self) -> None:
        self._known = self._ListClass()

    @property
    def category(self) -> int | None:
        """Veto category for this flag."""
        return self._category

    @category.setter
    def category(self, cat: int | None) -> None:
        self._category = if_not_none(int, cat)

    @property
    def description(self) -> str | None:
        """Description of why/how this flag was generated."""
        return self._description

    @description.setter
    def description(self, desc: str | None) -> None:
        self._description = desc

    @property
    def isgood(self) -> bool:
        """Whether `active` segments mean the instrument was in a good state."""
        return self._isgood

    @isgood.setter
    def isgood(self, good: bool) -> None:
        self._isgood = bool(good)

    @property
    def padding(self) -> tuple[float, float]:
        """(start, end) padding for this flag's active segments."""
        return self._padding

    @padding.setter
    def padding(self, pad: None | tuple[float | None, float | None]) -> None:
        if pad is None:
            self._padding = (0, 0)
        else:
            self._padding = (float(pad[0] or 0.), float(pad[1] or 0.))

    @padding.deleter
    def padding(self) -> None:
        self._padding = (0., 0.)

    # -- read-only properties --------

    @property
    def texname(self) -> str | None:
        """Name of this flag in LaTeX printable format."""
        if self.name is None:
            return None
        return self.name.replace("_", r"\_")

    @property
    def extent(self) -> Segment:
        """The GPS ``[start, stop)`` enclosing segment of this `DataQualityFlag`."""
        return self.known.extent()

    @property
    def livetime(self) -> float:
        """Amount of time this flag was `active`."""
        return abs(self.active)

    @property
    def regular(self) -> bool:
        """`True` if the `active` segments are a proper subset of the `known`."""
        return abs(self.active - self.known) == 0

    # -- i/o -------------------------

    read = UnifiedReadWriteMethod(DataQualityFlagRead)
    write = UnifiedReadWriteMethod(DataQualityFlagWrite)

    # -- classmethods ----------------

    @classmethod
    def query_dqsegdb(
        cls,
        flag: str,
        *args: SupportsToGps | Segment | SegmentList,
        host: str | None = DEFAULT_SEGMENT_SERVER,
        **kwargs,
    ) -> Self:
        """Query a DQSegDB server for a flag.

        Parameters
        ----------
        flag : `str`
            The name of the flag for which to query

        args
            Either, two `float`-like numbers indicating the
            GPS [start, stop) interval, or a `SegmentList`
            defining a number of summary segments

        host : `str`, optional
            Name or URL of the DQSegDB instance to talk to.
            Defaults to :func:`dqsegdb2.utils.get_default_host`.

        kwargs
            All other keyword arguments are passed to
            :func:`dqsegdb2.query.query_segments`.

        Returns
        -------
        flag : `DataQualityFlag`
            A new `DataQualityFlag`, with the `known` and `active` lists
            filled appropriately.

        Examples
        --------
        The GPS interval(s) of interest can be passed as two arguments
        specifing the start and end of a single interval:

        >>> DataQualityDict.query_dqsegdb(["X1:OBSERVING:1", "Y1:OBSERVING:1"], start, end)

        Or, as a single `Segment`:

        >>> DataQualityDict.query_dqsegdb(["X1:OBSERVING:1", "Y1:OBSERVING:1"], interval)

        Or, as a `SegmentList` specifying multiple intervals.

        >>> DataQualityDict.query_dqsegdb(["X1:OBSERVING:1", "Y1:OBSERVING:1"], intervals)
        """  # noqa: E501
        # parse arguments
        qsegs = _parse_query_segments(args, cls.query_dqsegdb)

        # parse deprecated 'url' keyword as 'host'
        url = kwargs.pop("url", None)
        if url:
            warnings.warn(
                "the `url` keyword argument for `query_dqsegdb` "
                "has been renamed `host`; this warning will become "
                "an error in the future",
                DeprecationWarning,
                stacklevel=2,
            )
            host = url

        # parse flag
        out = cls(name=flag)
        if out.ifo is None or out.tag is None:
            msg = f"Cannot parse ifo or tag (name) for flag '{flag}'"
            raise ValueError(msg)

        # process query
        for start, end in qsegs:
            # handle infinities
            if float(end) == +inf:
                end = int(to_gps("now"))  # noqa: PLW2901

            # query
            try:
                data = query_segments(
                    flag,
                    int(start),
                    int(end),
                    host=host,
                    **kwargs,
                )
            except HTTPError as exc:
                if exc.response.status_code == http_codes.NOT_FOUND:
                    exc.args = (exc.args[0] + f" [{flag}]",)
                raise

            # read from json buffer
            new = cls.read(
                BytesIO(json.dumps(data).encode("utf-8")),
                format="json",
            )

            # restrict to query segments
            segl = SegmentList([Segment(start, end)])
            new.known &= segl
            new.active &= segl
            out += new
            # replace metadata
            out.description = new.description
            out.isgood = new.isgood

        return out

    # alias for compatibility
    query = query_dqsegdb

    @classmethod
    def fetch_open_data(
        cls,
        flag: str,
        start: SupportsToGps,
        end: SupportsToGps,
        **kwargs,
    ) -> Self:
        """Fetch Open Data timeline segments into a flag.

        flag : `str`
            The name of the flag to query

        start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
            GPS start time of required data,
            any input parseable by `~gwpy.time.to_gps` is fine

        end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
            GPS end time of required data,
            any input parseable by `~gwpy.time.to_gps` is fine

        timeout : `int`, optional
            Timeout for download (seconds).

        host : `str`, optional
            URL of GWOSC host, default: ``'https://gwosc.org'``.

        Returns
        -------
        flag : `DataQualityFlag`
            a new flag with `active` segments filled from Open Data

        Examples
        --------
        >>> from gwpy.segments import DataQualityFlag
        >>> print(DataQualityFlag.fetch_open_data(
        ...     "H1_DATA",
        ...     "Sep 14 2015",
        ...     "Sep 15 2015",
        ... ))
        <DataQualityFlag('H1:DATA',
                         known=[[1126224017 ... 1126310417)]
                         active=[[1126251604 ... 1126252133)
                                 [1126252291 ... 1126274322)
                                 [1126276234 ... 1126281754)
                               ...
                                 [1126308670 ... 1126309577)
                                 [1126309637 ... 1126309817)
                                 [1126309877 ... 1126310417)]
                         description=None)>
        """
        start = int(to_gps(start))
        end = ceil(to_gps(end))
        known = [(start, end)]
        active = timeline.get_segments(flag, start, end, **kwargs)
        return cls(
            flag.replace("_", ":", 1),
            known=known,
            active=active,
            label=flag,
        )

    @classmethod
    def from_veto_def(
        cls,
        veto: igwn_ligolw.lsctables.VetoDef | astropy.table.Row,
    ) -> Self:
        """Define a `DataQualityFlag` from a `VetoDef`.

        Parameters
        ----------
        veto : `~igwn_ligolw.lsctables.VetoDef`, `~astropy.table.Row`.
            Veto definition to convert from.
        """
        # handle getting by item name (astropy) or attribute name (igwn_ligolw)
        if isinstance(veto, AstropyTableRow):
            get = veto.get
        else:
            def get(name, default=None):  # noqa: ANN001,ANN202
                return getattr(veto, name, default)

        name = f"{get('ifo')}:{get('name')}"
        with contextlib.suppress(TypeError):
            name += f":{int(get('version'))}"
        known = Segment(get("start_time"), get("end_time") or +inf)
        pad = (get("start_pad") or 0, get("end_pad") or 0)
        return cls(
            name=name,
            known=[known],
            category=get("category", None),
            description=get("comment", None),
            padding=pad,
        )

    # -- methods ---------------------

    def populate(
        self,
        source: str | None = DEFAULT_SEGMENT_SERVER,
        segments: SegmentList | None = None,
        *,
        pad: bool = True,
        **kwargs,
    ) -> Self:
        """Query the segment database for this flag's active segments.

        This method assumes all of the metadata for each flag have been
        filled. Minimally, the following attributes must be filled

        .. autosummary::

           ~DataQualityFlag.name
           ~DataQualityFlag.known

        Segments will be fetched from the database, with any
        :attr:`~DataQualityFlag.padding` added on-the-fly.

        This `DataQualityFlag` will be modified in-place.

        Parameters
        ----------
        source : `str`
            Source of segments for this flag. This must be
            either a URL for a segment database or a path to a file on disk.

        segments : `SegmentList`, optional
            A list of segments during which to query, if not given,
            existing known segments for this flag will be used.

        pad : `bool`, optional, default: `True`
            apply the `~DataQualityFlag.padding` associated with this
            flag, default: `True`.

        **kwargs
            any other keyword arguments to be passed to
            :meth:`DataQualityFlag.query` or :meth:`DataQualityFlag.read`.

        Returns
        -------
        self : `DataQualityFlag`
            a reference to this flag
        """
        tmp = DataQualityDict()
        tmp[self.name] = self
        tmp.populate(source=source, segments=segments, pad=pad, **kwargs)
        return tmp[self.name]

    def contract(self, x: float) -> SegmentList:
        """Contract each of the ``active`` segments by ``x`` seconds.

        This method adds ``x`` to each segment's lower bound, and subtracts
        ``x`` from the upper bound.

        The :attr:`~DataQualityFlag.active` `SegmentList` is modified
        in place.

        Parameters
        ----------
        x : `float`
            Number of seconds by which to contract each `Segment`.
        """
        self.active = self.active.contract(x)
        return self.active

    def protract(self, x: float) -> SegmentList:
        """Protract each of the ``active`` segments by ``x`` seconds.

        This method subtracts ``x`` from each segment's lower bound,
        and adds ``x`` to the upper bound, while maintaining that each
        `Segment` stays within the `known` bounds.

        The :attr:`~DataQualityFlag.active` `SegmentList` is modified
        in place.

        Parameters
        ----------
        x : `float`
            Number of seconds by which to protact each `Segment`.
        """
        self.active = self.active.protract(x)
        return self.active

    def pad(
        self,
        *args: float,
        inplace: bool = False,
    ) -> Self:
        """Apply a padding to each segment in this `DataQualityFlag`.

        This method either takes no arguments, in which case the value of
        the :attr:`~DataQualityFlag.padding` attribute will be used,
        or two values representing the padding for the start and end of
        each segment.

        For both the `start` and `end` paddings, a positive value means
        pad forward in time, so that a positive `start` pad or negative
        `end` padding will contract a segment at one or both ends,
        and vice-versa.

        This method will apply the same padding to both the
        `~DataQualityFlag.known` and `~DataQualityFlag.active` lists,
        but will not :meth:`~DataQualityFlag.coalesce` the result.

        Parameters
        ----------
        args : `float`, optional
            Two floats giving the start and end padding to apply.
            If not given, `self.padding` will be used.

        inplace : `bool`, optional
            Modify this object in-place.
            Default is `False`, i.e. return a copy of the original
            object with padded segments.

        Returns
        -------
        paddedflag : `DataQualityFlag`
            A view of the modified flag.
        """
        if not args:
            start, end = self.padding
        else:
            start, end = args

        if inplace:
            new = self
        else:
            new = self.copy()

        def _pad(s):
            return type(s)(s[0]+start, s[1]+end)

        new.known = type(self.known)(map(_pad, self.known))
        new.active = type(self.active)(map(_pad, self.active))

        return new

    def round(self, *, contract: bool = False) -> Self:
        """Round this flag to integer segments.

        Parameters
        ----------
        contract : `bool`, optional
            If `False` (default) expand each segment to the containing
            integer boundaries, otherwise contract each segment to the
            contained boundaries.

        Returns
        -------
        roundedflag : `DataQualityFlag`
            A copy of the original flag with the `active` and `known` segments
            padded out to integer boundaries..
        """
        def _round(seg: Segment) -> Segment:
            if contract:  # round inwards
                a = type(seg[0])(ceil(seg[0]))
                b = type(seg[1])(floor(seg[1]))
            else:  # round outwards
                a = type(seg[0])(floor(seg[0]))
                b = type(seg[1])(ceil(seg[1]))
            if a >= b:  # if segment is too short, return 'null' segment
                return type(seg)(0, 0)  # will get coalesced away
            return type(seg)(a, b)

        new = self.copy()
        new.active = type(new.active)(map(_round, new.active))
        new.known = type(new.known)(map(_round, new.known))
        return new.coalesce()

    def coalesce(self) -> Self:
        """Coalesce the segments for this flag.

        This method does two things:

        - `coalesces <SegmentList.coalesce>` the `~DataQualityFlag.known` and
          `~DataQualityFlag.active` segment lists
        - forces the `active` segments to be a proper subset of the `known`
          segments

        .. note::

            This operation is performed in-place.

        Returns
        -------
        self
            A view of this flag, not a copy.
        """
        self.known = self.known.coalesce()
        self.active = self.active.coalesce()
        self.active = (self.known & self.active).coalesce()
        return self

    def __repr__(self) -> str:
        """Return a representation of this flag."""
        prefix = f"<{type(self).__name__}("
        suffix = ")>"
        indent = " " * len(prefix)

        # format segment lists
        parts = {}
        for attr in ("known", "active"):
            parts[attr] = rep = textwrap.indent(
                str(getattr(self, attr)),
                f"{indent}      ",
            ).strip().splitlines()
            if len(rep) > 10:  # use ellipsis
                rep = [*rep[:3], f"{indent}      ...", *rep[-3:]]

        # print the thing
        return "".join((
            prefix,
            f"{os.linesep}{indent}".join([
                f"'{self.name}',",
                f"known={os.linesep.join(parts['known'])}",
                f"active={os.linesep.join(parts['active'])}",
                f"description='{self.description}'",
            ]),
            suffix,
        ))

    def copy(self) -> Self:
        """Build an exact copy of this flag.

        Returns
        -------
        flag : `DataQualityFlag`
            A deepcopy of the original flag.
        """
        return deepcopy(self)

    def plot(
        self,
        figsize: tuple[float, float] = (12, 4),
        xscale: str = "auto-gps",
        **kwargs,
    ) -> Plot:
        """Plot this flag on a segments projection.

        Parameters
        ----------
        figsize : `tuple` of `float`
            The size (width, height) of the figure to create.

        xscale: `str`
            The scaling to use for the X-axis (time axis).
            Default is ``"auto-gps"`` to dynamically choose the right scaling
            based on how much time is covered by the visible span.

        kwargs
            Other keyword arguments are passed to the
            :class:`~gwpy.plot.Plot` constructor.

        Returns
        -------
        figure : `~matplotlib.figure.Figure`
            The newly created figure, with populated Axes.

        See Also
        --------
        matplotlib.pyplot.figure
            For documentation of keyword arguments used to create the figure.

        matplotlib.figure.Figure.add_subplot
            For documentation of keyword arguments used to create the axes

        gwpy.plot.SegmentAxes.plot_segmentlist
            For documentation of keyword arguments used in rendering the data.
        """
        from matplotlib import rcParams

        from ..plot import Plot

        # get the right default label
        if "label" not in kwargs:
            if self.label:
                kwargs["label"] =  self.label
            elif rcParams["text.usetex"]:
                kwargs["label"] =  self.texname
            else:
                kwargs["label"] =  self.name

        return Plot(
            self,
            projection="segments",
            figsize=figsize,
            xscale=xscale,
            **kwargs,
        )

    def _parse_name(
        self,
        name: str | None,
    ) -> tuple[str | None, str | None, str | None]:
        """Parse a flag name and set properties of this flag.

        Parameters
        ----------
        name : `str`, `None`
            The full name of a `DataQualityFlag` to parse, e.g.
            ``"H1:DMT-SCIENCE:1"``, or `None` to set all components
            to `None`.

        Returns
        -------
        ifo : `str` or `None`
            The IFO prefix for this flag.

        tag : `str` or `None`
            The name tag for this flag.

        version : `int` or `None`
            The version number for this flag.

        Raises
        ------
        ValueError
            If the input ``name`` cannot be parsed into
            {ifo}:{tag}:{version} format.
        """
        if name is None:
            self.ifo = None
            self.tag = None
            self.version = None
        elif match := IFO_TAG_VERSION_REGEX.match(name):
            self.ifo = match["ifo"]
            self.tag = match["tag"]
            self.version = int(match["version"])
        elif match := IFO_TAG_REGEX.match(name):
            self.ifo = match["ifo"]
            self.tag = match["tag"]
            self.version = None
        elif match := TAG_VERSION_REGEX.match(name):
            self.ifo = None
            self.tag = match["tag"]
            self.version = int(match["version"])
        else:
            msg = (
                f"no flag name structure detected in '{name}', flags "
                "should be named as '{ifo}:{tag}:{version}'; "
                "for arbitrary strings, use the "
                "`DataQualityFlag.label` attribute"
            )
            raise ValueError(msg)
        return self.ifo, self.tag, self.version

    def __and__(self, other: Self) -> Self:
        """Find the intersection of this one and ``other``."""
        return self.copy().__iand__(other)

    def __iand__(self, other: Self) -> Self:
        """Intersect this flag with ``other`` in-place."""
        self.known &= other.known
        self.active &= other.active
        return self

    def __sub__(self, other: Self) -> Self:
        """Find the difference between this flag and another."""
        return self.copy().__isub__(other)

    def __isub__(self, other: Self) -> Self:
        """Subtract the ``other`` `DataQualityFlag` from this one in-place."""
        self.known &= other.known
        self.active -= other.active
        self.active &= self.known
        return self

    def __or__(self, other: Self) -> Self:
        """Find the union of this flag and ``other``."""
        return self.copy().__ior__(other)

    def __ior__(self, other: Self) -> Self:
        """Add the ``other`` `DataQualityFlag` to this one in-place."""
        self.known |= other.known
        self.active |= other.active
        return self

    __add__ = __or__
    __iadd__ = __ior__

    def __xor__(self, other: Self) -> Self:
        """Find the exclusive OR of this one and ``other``."""
        return self.copy().__ixor__(other)

    def __ixor__(self, other: Self) -> Self:
        """Exclusive OR this flag with ``other`` in-place."""
        self.known &= other.known
        self.active ^= other.active
        return self

    def __invert__(self) -> Self:
        """Return the logical inverse of this flag."""
        new = self.copy()
        new.active = ~self.active
        new.active &= new.known
        return new


class DataQualityDict(dict):
    """An `dict` of (key, `DataQualityFlag`) pairs."""

    _EntryClass: type[DataQualityFlag] = DataQualityFlag

    # -- i/o -------------------------

    read = UnifiedReadWriteMethod(DataQualityDictRead)
    write = UnifiedReadWriteMethod(DataQualityDictWrite)

    # -- classmethods ----------------

    @classmethod
    @_check_on_error
    def query_dqsegdb(
        cls,
        flags: list[str],
        *args: SupportsToGps | Segment | SegmentList,
        host: str | None = DEFAULT_SEGMENT_SERVER,
        on_error: str = "raise",
        parallel: int = 10,
        **kwargs,
    ) -> Self:
        """Query the advanced LIGO DQSegDB for a list of flags.

        Parameters
        ----------
        flags : `iterable`
            A list of flag names for which to query.

        args
            Either, two `float`-like numbers indicating the
            GPS [start, stop) interval, or a `SegmentList`
            defining a number of summary segments.

        host : `str`, optional
            Name or URL of the DQSegDB instance to talk to.
            Defaults to :func:`dqsegdb2.utils.get_default_host`.

        on_error : `str`, optional
            how to handle an error querying for one flag, one of

            - `'raise'` (default): raise the Exception
            - `'warn'`: print a warning
            - `'ignore'`: move onto the next flag as if nothing happened

        parallel : `int`, optional
            Maximum number of threads to use for parallel connections to
            the DQSegDB host.

        kwargs
            All other keyword arguments are passed to
            :func:`dqsegdb2.query.query_segments`.

        Returns
        -------
        flagdict : `DataQualityDict`
            An ordered `DataQualityDict` of (name, `DataQualityFlag`)
            pairs.

        Examples
        --------
        The GPS interval(s) of interest can be passed as two arguments
        specifing the start and end of a single interval:

        >>> DataQualityDict.query_dqsegdb(["X1:OBSERVING:1", "Y1:OBSERVING:1"], start, end)

        Or, as a single `Segment`:

        >>> DataQualityDict.query_dqsegdb(["X1:OBSERVING:1", "Y1:OBSERVING:1"], interval)

        Or, as a `SegmentList` specifying multiple intervals.

        >>> DataQualityDict.query_dqsegdb(["X1:OBSERVING:1", "Y1:OBSERVING:1"], intervals)
        """  # noqa: E501
        # parse segments
        qsegs = _parse_query_segments(args, cls.query_dqsegdb)

        # thread function
        def _query(flag: str) -> DataQualityFlag | None:
            try:
                return cls._EntryClass.query_dqsegdb(
                    flag,
                    qsegs,
                    host=host,
                    **kwargs,
                )
            except Exception as exc:
                exc.args = (f"{exc} [{flag}]",)
                if on_error == "raise":
                    raise
                if on_error == "warn":
                    warnings.warn(str(exc), stacklevel=2)
                return None

        # execute queries in threads
        out = cls()
        parallel = min(parallel, len(flags))
        with ThreadPoolExecutor(
            max_workers=parallel,
            thread_name_prefix=f"{cls.__name__}.query_dqsegdb",
        ) as pool:
            for flag in filter(None, pool.map(_query, flags)):
                out[flag.name] = flag

        return out

    # alias for compatibility
    query = query_dqsegdb

    @classmethod
    def from_veto_definer_file(
        cls,
        source: str,
        start: SupportsToGps | None = None,
        end: SupportsToGps | None = None,
        ifo: str | None = None,
        **read_kw,
    ) -> Self:
        """Read a `DataQualityDict` from a LIGO_LW XML VetoDefinerTable.

        Parameters
        ----------
        source : `str`, `Path`, `file`
            Path or URL of veto definer file to read, or open file.

        start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
            GPS start time of required data,
            any input parseable by `~gwpy.time.to_gps` is fine

        end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
            GPS end time of required data,
            any input parseable by `~gwpy.time.to_gps` is fine

        ifo : `str`, optional
            Interferometer prefix whose flags you want to read.
            Default is `None` (read all flags).

        read_kw
            Other keyword arguments are passed to
            `~astropy.table.Table.read` when reading the veto definer file.

        Returns
        -------
        flags : `DataQualityDict`
            A `DataQualityDict` of flags parsed from the `veto_def_table`
            of the input file.

        Notes
        -----
        This method does not automatically `~DataQualityDict.populate`
        the `active` segment list of any flags, a separate call should
        be made for that as follows

        >>> flags = DataQualityDict.from_veto_definer_file('/path/to/file.xml')
        >>> flags.populate()
        """
        from ..table import Table

        # read veto definer file as a table
        read_kw.setdefault("tablename", "veto_definer")
        read_kw.setdefault("columns", [
            "ifo",
            "name",
            "version",
            "category",
            "comment",
            "start_time",
            "start_pad",
            "end_time",
            "end_pad",
        ])
        veto_def_table = Table.read(source, **read_kw)

        # handle GPS types
        startgps = None if start is None else to_gps(start)
        endgps = None if end is None else to_gps(end)

        def _keep(row: astropy.table.Row) -> bool:
            """Return `True` if this row is relevant."""
            if ifo and row["ifo"] != ifo:
                return False
            if startgps and 0 < row["end_time"] <= startgps:
                return False
            return not (endgps and row["start_time"] >= endgps)

        # parse flag definitions
        out = cls()
        for row in filter(_keep, veto_def_table):
            # pin the times
            if startgps:
                row["start_time"] = max(row["start_time"], startgps)
            if endgps and not row["end_time"]:
                row["end_time"] = endgps
            elif endgps:
                row["end_time"] = min(row["end_time"], endgps)
            flag = DataQualityFlag.from_veto_def(row)
            if flag.name in out:
                out[flag.name] |= flag
            else:
                out[flag.name] = flag
        return out

    @classmethod
    def from_ligolw_tables(
        cls,
        segmentdeftable: igwn_ligolw.lsctables.SegmentDefTable,
        segmentsumtable: igwn_ligolw.lsctables.SegmentSumTable,
        segmenttable: igwn_ligolw.lsctables.SegmentTable,
        names: list[str] | None = None,
        gpstype: type[SupportsFloat] | Callable[[float], SupportsFloat] = LIGOTimeGPS,
        on_missing: str = "error",
    ) -> Self:
        """Build a `DataQualityDict` from a set of LIGO_LW segment tables.

        Parameters
        ----------
        segmentdeftable : :class:`~igwn_ligolw.lsctables.SegmentDefTable`
            The ``segment_definer`` table to read.

        segmentsumtable : :class:`~igwn_ligolw.lsctables.SegmentSumTable`
            The ``segment_summary`` table to read.

        segmenttable : :class:`~igwn_ligolw.lsctables.SegmentTable`
            The ``segment`` table to read.

        names : `list` of `str`, optional
            A list of flag names to read.
            Default is to read all names.

        gpstype : `type`, `callable`, optional
            Class to use for GPS times in returned objects, can be a function
            to convert GPS time to something else.
            Default is `~gwpy.time.LIGOTimeGPS`.

        on_missing : `str`, optional
            Action to take when a one or more ``names`` are not found in
            the ``segment_definer`` table, one of

            - ``'ignore'`` : do nothing
            - ``'warn'`` : print a warning
            - ``error'`` : raise a `ValueError`

        Returns
        -------
        dqdict : `DataQualityDict`
            A dict of `DataQualityFlag` objects populated from the LIGO_LW tables.
        """
        out = cls()
        id_: dict[str, list[int]] = {}  # need to record relative IDs from LIGO_LW

        # read segment definers and generate DataQualityFlag object
        for row in segmentdeftable:
            ifos = sorted(row.instruments)
            ifo = "".join(ifos) if ifos else None
            tag = row.name
            version = row.version
            name = ":".join(
                str(k) for k in (ifo, tag, version) if k is not None
            )
            if names is None or name in names:
                out[name] = DataQualityFlag(name)
                thisid = int(row.segment_def_id)
                try:
                    id_[name].append(thisid)
                except KeyError:
                    id_[name] = [thisid]

        # verify all requested flags were found
        for flag in names or []:
            if flag not in out and on_missing != "ignore":
                msg = f"no segment definition found for flag='{flag}' in file"
                if on_missing == "warn":
                    warnings.warn(msg, stacklevel=2)
                else:
                    raise ValueError(msg)


        def _parse_segments(
            table: igwn_ligolw.ligolw.Table,
            listattr: str,
        ) -> None:
            """Parse a table into the target DataQualityDict."""
            # handle missing *_ns columns in LIGO_LW XML
            # (LIGO DMT doesn't/didn't write them)
            if "start_time_ns" in table.columnnames:
                row_segment = operator.attrgetter("segment")
            else:
                row_segment = operator.attrgetter("start_time", "end_time")

            for row in table:
                for flag in out:
                    # match row ID to list of IDs found for this flag
                    if int(row.segment_def_id) in id_[flag]:
                        getattr(out[flag], listattr).append(
                            Segment(*map(gpstype, row_segment(row))),
                        )
                        break

        # read segment summary table as 'known'
        _parse_segments(segmentsumtable, "known")

        # read segment table as 'active'
        _parse_segments(segmenttable, "active")

        return out

    def to_ligolw_tables(
        self,
        **attrs,
    ) -> tuple[
        igwn_ligolw.lsctables.SegmentDefTable,
        igwn_ligolw.lsctables.SegmentSumTable,
        igwn_ligolw.lsctables.SegmentTable,
    ]:
        """Convert this `DataQualityDict` into a trio of LIGO_LW segment tables.

        Parameters
        ----------
        attrs
            Other attributes to add to all rows in all tables
            (e.g. ``'process_id'``).

        Returns
        -------
        segmentdeftable : :class:`~igwn_ligolw.lsctables.SegmentDefTable`
            The ``segment_definer`` table.

        segmentsumtable : :class:`~igwn_ligolw.lsctables.SegmentSumTable`
            The ``segment_summary`` table.

        segmenttable : :class:`~igwn_ligolw.lsctables.SegmentTable`
            The ``segment`` table.
        """
        from igwn_ligolw import lsctables

        from ..io.ligolw import to_table_type as to_ligolw_table_type

        segdeftab_class = lsctables.SegmentDefTable
        segsumtab_class = lsctables.SegmentSumTable
        segtab_class = lsctables.SegmentTable
        segdeftab = segdeftab_class.new()
        segsumtab = segsumtab_class.new()
        segtab = segtab_class.new()

        def _write_attrs(
            table: igwn_ligolw.ligolw.Table,
            row: igwn_ligolw.ligolw.Row,
        ) -> None:
            """Write custom attributes to this row."""
            for key, val in attrs.items():
                setattr(row, key, to_ligolw_table_type(val, table, key))

        # write flags to tables
        for flag in self.values():
            # segment definer
            segdef = segdeftab.RowType()
            for col in segdeftab.columnnames:  # default all columns to None
                setattr(segdef, col, None)
            segdef.instruments = {flag.ifo}
            segdef.name = flag.tag
            segdef.version = flag.version
            segdef.comment = flag.description
            segdef.insertion_time = to_gps("now").gpsSeconds
            segdef.segment_def_id = segdeftab_class.get_next_id()
            _write_attrs(segdeftab, segdef)
            segdeftab.append(segdef)

            # write segment summary (known segments)
            for vseg in flag.known:
                segsum = segsumtab.RowType()
                for col in segsumtab.columnnames:  # default columns to None
                    setattr(segsum, col, None)
                segsum.segment_def_id = segdef.segment_def_id
                segsum.segment = map(LIGOTimeGPS, vseg)
                segsum.comment = None
                segsum.segment_sum_id = segsumtab_class.get_next_id()
                _write_attrs(segsumtab, segsum)
                segsumtab.append(segsum)

            # write segment table (active segments)
            for aseg in flag.active:
                seg = segtab.RowType()
                for col in segtab.columnnames:  # default all columns to None
                    setattr(seg, col, None)
                seg.segment_def_id = segdef.segment_def_id
                seg.segment = map(LIGOTimeGPS, aseg)
                seg.segment_id = segtab_class.get_next_id()
                _write_attrs(segtab, seg)
                segtab.append(seg)

        return segdeftab, segsumtab, segtab

    # -- methods ---------------------

    def coalesce(self) -> Self:
        """Coalesce all segments lists in this `DataQualityDict`.

        **This method modifies this object in-place.**

        Returns
        -------
        self
            A view of this flag, not a copy.
        """
        for flag in self:
            self[flag].coalesce()
        return self

    @_check_on_error
    def populate(
        self,
        source: str | None = DEFAULT_SEGMENT_SERVER,
        segments: SegmentList | None = None,
        *,
        pad: bool = True,
        on_error: str = "raise",
        **kwargs,
    ) -> Self:
        """Query the segment database for each flag's active segments.

        This method assumes all of the metadata for each flag have been
        filled. Minimally, the following attributes must be filled

        .. autosummary::

           ~DataQualityFlag.name
           ~DataQualityFlag.known

        Segments will be fetched from the database, with any
        :attr:`~DataQualityFlag.padding` added on-the-fly.

        Entries in this dict will be modified in-place.

        Parameters
        ----------
        source : `str`
            Source of segments for this flag. This must be
            either a URL for a segment database or a path to a file on disk.

        segments : `SegmentList`, optional
            A list of known segments during which to query, if not given,
            existing known segments for flags will be used.

        pad : `bool`, optional, default: `True`
            Apply the `~DataQualityFlag.padding` associated with each
            flag, default: `True`.

        on_error : `str`
            How to handle an error querying for one flag, one of

            - `'raise'` (default): raise the Exception
            - `'warn'`: print a warning
            - `'ignore'`: move onto the next flag as if nothing happened

        kwargs
            Any other keyword arguments to be passed to
            :meth:`DataQualityFlag.query` or :meth:`DataQualityFlag.read`.

        Returns
        -------
        self : `DataQualityDict`
            A reference to the modified DataQualityDict
        """
        # format source
        url = urlparse(source)
        # query DQSegDB if the URL is just a hostname
        use_dqsegdb = bool(url.netloc and url.path in ("", "/"))

        # if given the segments to use, query DQSegDB for everthing now
        if not use_dqsegdb:
            tmp = type(self).read(source, **kwargs)
        elif segments is not None:
            segments = SegmentList(map(Segment, segments))
            tmp = type(self).query_dqsegdb(
                self.keys(),
                segments,
                host=source,
                on_error=on_error,
                **kwargs,
            )

        # apply padding and wrap to given known segments
        for key in self:
            if use_dqsegdb and segments is None:
                # get segments for this flag now
                try:
                    tmp = {key: self[key].query_dqsegdb(
                        self[key].name,
                        self[key].known,
                        host=source,
                        **kwargs,
                    )}
                except RequestException as exc:
                    if on_error == "raise":
                        raise
                    if on_error == "warn":
                        warnings.warn(
                            f"Error querying for '{key}': {exc}",
                            stacklevel=2,
                        )
                    continue
            self[key].known &= tmp[key].known
            self[key].active = tmp[key].active
            if pad:
                self[key] = self[key].pad(inplace=True)
                if segments is not None:
                    self[key].known &= segments
                    self[key].active &= segments
        return self

    def copy(self, *, deep: bool = False) -> Self:
        """Build a copy of this dictionary.

        Parameters
        ----------
        deep : `bool`, optional, default: `False`
            perform a deep copy of the original dictionary with a fresh
            memory address

        Returns
        -------
        flag2 : `DataQualityFlag`
            a copy of the original dictionary
        """
        if deep:
            return deepcopy(self)
        return shallowcopy(self)

    def __iand__(self, other: Self) -> Self:
        """Intersect this dict with ``other`` in-place."""
        for key, value in other.items():
            if key in self:
                self[key] &= value
            else:
                self[key] = self._EntryClass()
        return self

    def __and__(self, other: Self) -> Self:
        """Find the intersection of this dict and ``other``."""
        if (
            sum(len(s.active) for s in self.values())
            <= sum(len(s.active) for s in other.values())
        ):
            return self.copy(deep=True).__iand__(other)
        return other.copy(deep=True).__iand__(self)

    def __ior__(self, other: Self) -> Self:
        """Add the ``other`` `DataQualityDict` to this one in-place."""
        for key, value in other.items():
            if key in self:
                self[key] |= value
            else:
                self[key] = shallowcopy(value)
        return self

    def __or__(self, other: Self) -> Self:
        """Find the union of this dict and ``other``."""
        if (
            sum(len(s.active) for s in self.values())
            >= sum(len(s.active) for s in other.values())
        ):
            return self.copy(deep=True).__ior__(other)
        return other.copy(deep=True).__ior__(self)

    __iadd__ = __ior__
    __add__ = __or__

    def __isub__(self, other: Self) -> Self:
        """Subtract the ``other`` `DataQualityDict` from this one in-place."""
        for key, value in other.items():
            if key in self:
                self[key] -= value
        return self

    def __sub__(self, other: Self) -> Self:
        """Find the difference between this dict and another."""
        return self.copy(deep=True).__isub__(other)

    def __ixor__(self, other: Self) -> Self:
        """Exclusive OR this dict with ``other`` in-place."""
        for key, value in other.items():
            if key in self:
                self[key] ^= value
        return self

    def __xor__(self, other: Self) -> Self:
        """Find the exclusive OR of this dict and ``other``."""
        return self.copy(deep=True).__ixor__(other)

    def __invert__(self) -> Self:
        """Return the logical inverse of this dict."""
        new = self.copy(deep=True)
        for key, value in new.items():
            new[key] = ~value
        return new

    def union(self) -> DataQualityFlag:
        """Return the union of all flags in this dict.

        Returns
        -------
        union : `DataQualityFlag`
            a new `DataQualityFlag` who's active and known segments
            are the union of those of the values of this dict
        """
        usegs = reduce(operator.or_, self.values())
        usegs.name = " | ".join(self.keys())
        return usegs

    def intersection(self) -> DataQualityFlag:
        """Return the intersection of all flags in this dict.

        Returns
        -------
        intersection : `DataQualityFlag`
            a new `DataQualityFlag` who's active and known segments
            are the intersection of those of the values of this dict
        """
        isegs = reduce(operator.and_, self.values())
        isegs.name = " & ".join(self.keys())
        return isegs

    def plot(
        self,
        label: str = "key",
        **kwargs,
    ) -> Plot:
        """Plot this dict on a segments projection.

        Parameters
        ----------
        label : `str`, optional
            Labelling system to use, or fixed label for all flags,
            special values include

            - ``'key'``: use the key of the `DataQualityDict`,
            - ``'name'``: use the :attr:`~DataQualityFlag.name` of the flag

            If anything else, that fixed label will be used for all lines.

        kwargs
            All keyword arguments are passed to the
            :class:`~gwpy.plot.Plot` constructor.

        Returns
        -------
        figure : `~matplotlib.figure.Figure`
            the newly created figure, with populated Axes.

        See Also
        --------
        matplotlib.pyplot.figure
            For documentation of keyword arguments used to create the figure.

        matplotlib.figure.Figure.add_subplot
            For documentation of keyword arguments used to create the axes.

        gwpy.plot.SegmentAxes.plot_segmentlist
            For documentation of keyword arguments used in rendering the data.
        """
        # make plot
        from ..plot import Plot
        return Plot(
            [self],
            projection="segments",
            label=label,
            **kwargs,
        )



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwpy.signal.filter_design`."""

from __future__ import annotations

from typing import TYPE_CHECKING
from unittest import mock

import numpy
import pytest
from astropy import units
from numpy.testing import assert_array_equal
from scipy import signal

from ...testing import utils
from .. import filter_design

if TYPE_CHECKING:
    from ..filter_design import (
        TapsType,
        ZpkCompatible,
        ZpkType,
    )

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

ONE_HZ = units.Quantity(1, "Hz")

# -- filters --------------------------

FILTER_FS = 1024.
FILTER_NYQ = FILTER_FS / 2.
NOTCH_F = 60.
LOWPASS_F = 100.
HIGHPASS_F = 200.
BANDPASS_F = (LOWPASS_F, HIGHPASS_F)


@pytest.fixture(scope="module")
def notch_60() -> ZpkType:
    """60 Hz notch filter fixture."""
    passband = (NOTCH_F - 1) / FILTER_NYQ, (NOTCH_F + 1) / FILTER_NYQ
    stopband = (NOTCH_F - .1) / FILTER_NYQ, (NOTCH_F + .1) / FILTER_NYQ
    return signal.iirdesign(
        passband,
        stopband,
        1,  # max passband loss (dB)
        10,  # min stopband loss (dB)
        analog=False,
        ftype="ellip",
        output="zpk",
    )


@pytest.fixture(scope="module")
def lowpass_100_iir() -> ZpkType:
    """100 Hz lowpass filter fixture."""
    return signal.iirdesign(
        LOWPASS_F / FILTER_NYQ,
        LOWPASS_F * 1.5 / FILTER_NYQ,
        2,
        30,
        analog=False,
        ftype="cheby1",
        output="zpk",
    )


@pytest.fixture(scope="module")
def lowpass_100_fir() -> TapsType:
    """100 Hz lowpass FIR filter fixture."""
    return signal.firwin(
        30,
        LOWPASS_F,
        window="hamming",
        width=50.,
        fs=FILTER_FS,
    )


@pytest.fixture(scope="module")
def highpass_100_iir() -> ZpkType:
    """100 Hz highpass filter fixture."""
    return signal.iirdesign(
        HIGHPASS_F / FILTER_NYQ,
        HIGHPASS_F * 2 / 3. / FILTER_NYQ,
        2,
        30,
        analog=False,
        ftype="cheby1",
        output="zpk",
    )


@pytest.fixture(scope="module")
def highpass_100_fir() -> TapsType:
    """100 Hz highpass FIR filter fixture."""
    return signal.firwin(
        23,
        HIGHPASS_F,
        window="hamming",
        pass_zero=False,
        width=-HIGHPASS_F / 3.,
        fs=FILTER_FS,
    )


@pytest.fixture(scope="module")
def bandpass_100_200_iir() -> ZpkType:
    """100-200 Hz bandpass FIR filter fixture."""
    return signal.iirdesign(
        (LOWPASS_F / FILTER_NYQ, HIGHPASS_F / FILTER_NYQ),
        (LOWPASS_F * 2 / 3. / FILTER_NYQ, HIGHPASS_F * 3 / 2. / FILTER_NYQ),
        2,
        30,
        analog=False,
        ftype="cheby1",
        output="zpk",
    )


@pytest.fixture(scope="module")
def bandpass_100_200_fir() -> TapsType:
    """100-200 Hz bandpass FIR filter fixture."""
    return signal.firwin(
        45,
        BANDPASS_F,
        window="hamming",
        pass_zero=False,
        fs=FILTER_FS,
    )


# -- tests ----------------------------

def test_truncate_transfer():
    """Test `_truncate_transfer()`."""
    series = numpy.ones(64)
    trunc = filter_design._truncate_transfer(series)
    assert trunc[0] == 0
    assert trunc[-1] == 0
    utils.assert_allclose(series[5:59], trunc[5:59])


def test_truncate_impulse():
    """Test `_truncate_impulse()`."""
    series = numpy.ones(64)
    trunc = filter_design._truncate_impulse(series, ntaps=10)
    assert trunc[0] != 0
    assert trunc[-1] != 0
    utils.assert_allclose(trunc[5:59], numpy.zeros(54))


def test_fir_from_transfer():
    """Test `fir_from_transfer()`."""
    frequencies = numpy.arange(0, 64)
    fseries = numpy.cos(2 * numpy.pi * frequencies)

    # prepare the time domain filter
    fir = filter_design.fir_from_transfer(fseries, ntaps=10)

    # test the filter
    assert abs(fir[0] / fir.max()) <= 1e-2
    assert abs(fir[-1] / fir.max()) <= 1e-2
    assert fir.size == 10


def test_notch_iir(notch_60):
    """Test `notch()` with an IIR filter."""
    # test simple notch
    zpk = filter_design.notch(NOTCH_F, FILTER_FS)
    utils.assert_zpk_equal(zpk, notch_60)


def test_notch_iir_quantities(notch_60):
    """Test `notch()` with an IIR filter."""
    # test Quantities
    zpk = filter_design.notch(NOTCH_F * ONE_HZ, FILTER_FS * ONE_HZ)
    utils.assert_zpk_equal(zpk, notch_60)


def test_notch_fir_notimplemented():
    """Test `notch()` with an FIR filter."""
    with pytest.raises(NotImplementedError):
        filter_design.notch(60, 16384, type="fir")  # type: ignore[call-overload]


def test_lowpass_iir(lowpass_100_iir):
    """Test `lowpass()` with an IIR filter."""
    iir = filter_design.lowpass(LOWPASS_F, FILTER_FS)
    utils.assert_zpk_equal(iir, lowpass_100_iir)


def test_lowpass_fir(lowpass_100_fir):
    """Test `lowpass()` with an FIR filter."""
    fir = filter_design.lowpass(LOWPASS_F, FILTER_FS, type="fir")
    utils.assert_allclose(fir, lowpass_100_fir)


def test_highpass_iir(highpass_100_iir):
    """Test `highpass()` with an IIR filter."""
    iir = filter_design.highpass(HIGHPASS_F, FILTER_FS)
    utils.assert_zpk_equal(iir, highpass_100_iir)


def test_highpass_fir(highpass_100_fir):
    """Test `highpass()` with an FIR filter."""
    fir = filter_design.highpass(HIGHPASS_F, FILTER_FS, type="fir")
    utils.assert_allclose(fir, highpass_100_fir)


def test_bandpass_iir(bandpass_100_200_iir):
    """Test `bandpass()` with an IIR filter."""
    iir = filter_design.bandpass(LOWPASS_F, HIGHPASS_F, FILTER_FS)
    utils.assert_zpk_equal(iir, bandpass_100_200_iir)


def test_bandpass_fir(bandpass_100_200_fir):
    """Test `bandpass()` with an FIR filter."""
    fir = filter_design.bandpass(LOWPASS_F, HIGHPASS_F, FILTER_FS, type="fir")
    utils.assert_allclose(fir, bandpass_100_200_fir)


def test_concatenate_zpks():
    """Test `concatenate_zpks()`."""
    asa = numpy.asarray
    z1, p1, k1 = [1, 2, 3], [4, 5, 6], 1.
    z2, p2, k2 = [1, 2, 3, 4], [5, 6, 7, 8], 100
    utils.assert_zpk_equal(
        filter_design.concatenate_zpks((z1, p1, k1), (z2, p2, k2)),
        (asa(z1 + z2), asa(p1 + p2), k1 * k2),
    )


def test_parse_filter_fir():
    """Test `parse_filter()` with an FIR filter."""
    taps = numpy.arange(10)
    assert filter_design.parse_filter(taps) == (
        "ba",
        (taps, [1.]),
    )


def test_parse_filter_ba():
    """Test :func:`parse_filter` with a transfer function filter."""
    in_ = signal.butter(1, 0.5, output="ba")
    typ, out = filter_design.parse_filter(in_)
    assert typ == "ba"
    assert_array_equal(in_[0], out[0])
    assert_array_equal(in_[1], out[1])


def test_parse_filter_zpk():
    """Test `parse_filter()` with a ZPK filter."""
    in_ = signal.butter(1, 0.5, output="zpk")
    typ, out = filter_design.parse_filter(in_)
    assert typ == "zpk"
    utils.assert_zpk_equal(in_, out)


def test_parse_filter_sos():
    """Test `parse_filter()` with a SOS filter."""
    in_ = signal.butter(4, 0.5, output="sos")
    typ, out = filter_design.parse_filter(in_)
    assert typ == "sos"
    assert_array_equal(in_, out)


def test_parse_filter_lti():
    """Test `parse_filter()` with an LTI filter."""
    in_ = signal.lti(*signal.butter(2, 0.5, output="zpk"))
    typ, out = filter_design.parse_filter(in_)
    assert typ == "zpk"
    utils.assert_zpk_equal(
        (in_.zeros, in_.poles, in_.gain),
        out,
    )


@pytest.fixture
def example_zpk_fs_tuple() -> tuple[ZpkCompatible, float]:
    """Return an example ZPK and sample rate."""
    z, p, k = [1], [2], 0.1
    fs = 0.1
    return (z, p, k), fs


def test_convert_to_digital_zpk(example_zpk_fs_tuple):
    """Test `convert_to_digital()` with ZPK input."""
    (z, p, k), fs = example_zpk_fs_tuple

    dform, dfilt = filter_design._convert_to_digital((z, p, k), fs)

    assert dform == "zpk"
    assert dfilt == signal.bilinear_zpk(z, p, k, fs)


def test_convert_to_digital_ba():
    """Test `convert_to_digital()` with BA input."""
    b, a = signal.butter(1, .5, output="ba")

    # this should be converted to ZPK form
    dform, dfilt = filter_design._convert_to_digital((b, a), 16)
    assert dform == "ba"
    dfilt_exp = signal.bilinear(b, a, fs=16)
    assert_array_equal(dfilt[0], dfilt_exp[0])
    assert_array_equal(dfilt[1], dfilt_exp[1])


def test_convert_to_digital_fir():
    """Test `convert_to_digital()` with FIR (numerator only) input."""
    fs = 0.1
    b = numpy.array([1, 0.2, 0.5])
    # this should be converted to ZPK form
    dform, dfilt = filter_design._convert_to_digital(b, fs)
    assert dform == "ba"
    assert numpy.allclose(dfilt, signal.bilinear(b, [1], fs))


def test_convert_to_digital_complex_type_preserved():
    """Test that conversion to digital does not erroneously convert to float types.

    Tests regression against:
    https://gitlab.com/gwpy/gwpy/-/issues/1630#note_2001485594
    """
    z, p, k = signal.butter(3, 30, "low", analog=True, output="zpk")
    form, filt = filter_design._convert_to_digital((z, p, k), 100)
    assert form == "zpk"
    assert p.dtype == filt[1].dtype
    assert numpy.iscomplexobj(filt[1])


@mock.patch(
    "gwpy.signal.filter_design.parse_filter",
    return_value=("invalid", [1, 2, 3]),
)
def test_convert_to_digital_invalid_form(mock_parse_filter):
    """Test that `convert_to_digital()` raises ValueError on invalid filter form."""
    with pytest.raises(
        ValueError,
        match="convert 'invalid'",
    ):
        filter_design._convert_to_digital([1, 2, 3], sample_rate=1)



# Copyright (c) 2018-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Environment utilities for GWpy."""

from __future__ import annotations

import os

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

TRUE: tuple[str, ...] = (
    "1",
    "y",
    "yes",
    "true",
)


def bool_env(
    key: str,
    default: bool = False,  # noqa: FBT001,FBT002
) -> bool:
    """Parse an environment variable as a boolean switch.

    `True` is returned if the variable value matches one of the following:

    - ``'1'``
    - ``'y'``
    - ``'yes'``
    - ``'true'``

    The match is case-insensitive (so ``'Yes'`` will match as `True`).

    Parameters
    ----------
    key : `str`
        The name of the environment variable to find.

    default : `bool`
        The default return value if the key is not found.

    Returns
    -------
    True
        If the environment variable matches as 'yes' or similar.
    False
        Otherwise.

    Examples
    --------
    >>> import os
    >>> os.environ['GWPY_VALUE'] = 'yes'
    >>> print(bool_env('GWPY_VALUE'))
    True
    >>> os.environ['GWPY_VALUE'] = 'something else'
    >>> print(bool_env('GWPY_VALUE'))
    False
    >>> print(bool_env('GWPY_VALUE2'))
    False
    """
    try:
        return os.environ[key].lower() in TRUE
    except KeyError:
        return default



# Copyright (c) 2018-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.plot`."""

import numpy
import pytest

from ...segments import (
    Segment,
    SegmentList,
)
from ...testing import utils
from ...types import (
    Array2D,
    Series,
)
from .. import Plot
from .utils import FigureTestBase

RNG = numpy.random.default_rng(seed=0)


# -- test classes --------------------

class TestPlot(FigureTestBase):
    """Tests of `gwpy.plot.Plot`."""

    FIGURE_CLASS = Plot

    def test_init(self):
        """Test `Plot()` initialisation."""
        plot = self.FIGURE_CLASS(figsize=(4, 3), dpi=100)
        try:
            assert tuple(plot.get_size_inches()) == (4., 3.)
        finally:
            plot.close()

    def test_init_empty(self):
        """Test empty `Plot()` initialisation."""
        plot = self.FIGURE_CLASS(geometry=(2, 2))
        try:
            assert len(plot.axes) == 4
            assert plot.axes[-1].get_subplotspec().get_geometry() == (2, 2, 3, 3)
        finally:
            plot.close()

    def test_init_with_data(self):
        """Test `Plot()` with data."""
        plot = self.FIGURE_CLASS([1, 2, 3, 4])
        try:
            assert len(plot.axes) == 1
            utils.assert_array_equal(
                plot.gca().lines[-1].get_ydata(),
                numpy.array([1, 2, 3, 4]),
            )
        finally:
            plot.close()

    def test_plot_series(self):
        """Test `Plot(<Series> ...)`."""
        a = Series(range(10), dx=.1)
        plot = self.FIGURE_CLASS(a)
        try:
            ax = plot.gca()
            line = ax.lines[0]
            assert len(plot.axes) == 1
            assert len(ax.lines) == 1
            utils.assert_array_equal(line.get_xdata(), a.xindex.value)
            utils.assert_array_equal(line.get_ydata(), a.value)
        finally:
            plot.close()

    def test_plot_multiple_series(self):
        """Test `Plot(<Series>, <Series>, ...)`."""
        a = Series(range(10), dx=.1)
        b = Series(range(10), dx=.1)
        plot = self.FIGURE_CLASS(a, b)
        try:
            assert len(plot.axes) == 1
            assert len(plot.axes[0].lines) == 2
        finally:
            plot.close()

    def test_plot_multiple_series_separate(self):
        """Test `Plot(<Series>, <Series>, separate=True), ...)`."""
        a = Series(range(10), dx=.1)
        b = Series(range(10), dx=.1)
        plot = self.FIGURE_CLASS(a, b, separate=True, sharex=True, sharey=True)
        try:
            assert len(plot.axes) == 2
            for i, ax in enumerate(plot.axes):
                assert ax.get_subplotspec().get_geometry() == (2, 1, i, i)
                assert len(ax.lines) == 1
            assert plot.axes[1]._sharex is plot.axes[0]
        finally:
            plot.close()

    def test_plot_array2d_imshow(self):
        """Test `Plot(<Array2D>, method="imshow")`."""
        array = Array2D(RNG.random(size=(10, 10)), dx=.1, dy=.2)
        plot = self.FIGURE_CLASS(array, method="imshow")
        try:
            assert len(plot.axes[0].images) == 1
            image = plot.axes[0].images[0]
            utils.assert_array_equal(image.get_array(), array.value.T)
        finally:
            plot.close()

    def test_save(self, fig, tmp_path):
        """Test `Plot.save`."""
        tmp = tmp_path / "plot.png"
        fig.save(str(tmp))
        assert tmp.is_file()

    def test_get_axes(self, fig):
        """Test `Plot.get_axes`."""
        fig.add_subplot(2, 1, 1, projection="rectilinear")
        fig.add_subplot(2, 1, 2, projection="polar")
        assert fig.get_axes() == fig.axes
        assert fig.get_axes(projection="polar") == fig.axes[1:]

    def test_colorbar(self, fig):
        """Test `Plot.colorbar`."""
        ax = fig.gca()
        array = Array2D(RNG.random(size=(10, 10)), dx=.1, dy=.2)
        image = ax.imshow(array)
        cbar = fig.colorbar(vmin=2, vmax=4, fraction=0.)
        assert cbar.mappable is image
        assert cbar.mappable.get_clim() == (2., 4.)

    def test_add_segments_bar(self, fig):
        """Test `Plot.add_segments_bar`."""
        ax = fig.add_subplot(xscale="auto-gps", epoch=150)
        ax.set_xlim(100, 200)
        ax.set_xlabel("test")
        segs = SegmentList([Segment(10, 110), Segment(150, 400)])
        segax = fig.add_segments_bar(segs)
        assert segax._sharex is ax
        assert ax.get_xlabel() == ""
        for ax_ in (ax, segax):
            assert ax_.get_xlim() == (100, 200)
            assert ax_.get_epoch() == 150.

        # check that it works again
        segax = fig.add_segments_bar(segs, ax=ax)

    def test_add_segments_bar_location_error(self, fig):
        """Test `Plot.add_segments_bar` error handling."""
        with pytest.raises(
            ValueError,
            match="segments can only be positoned at 'top' or 'bottom'",
        ):
            fig.add_segments_bar([], location="left")



# Copyright (c) 2018-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for testing `gwpy.plot`."""

from __future__ import annotations

from io import BytesIO

import pytest
from matplotlib import pyplot
from matplotlib.axes import Axes

from .. import Plot


@pytest.mark.usefixtures("usetex")
class _Base:
    """Base class for tests of `Plot` and `Axes`."""

    @staticmethod
    def save(fig):
        """Save a figure to a 'file' in memory."""
        out = BytesIO()
        fig.savefig(out)
        return fig

    @classmethod
    def save_and_close(cls, fig):
        """Save a figure to a 'file' in memory and then close it."""
        cls.save(fig)
        try:
            fig.close()
        except AttributeError:
            pyplot.close(fig)
        return fig


class FigureTestBase(_Base):
    """Base class for tests of `Plot`."""

    FIGURE_CLASS: type[Plot] = Plot

    @pytest.fixture
    @classmethod
    def fig(cls):
        """Yield a new instance of `.FIGURE_CLASS`.

        This fixture checks that the figure can be rendered as a PNG (in memory)
        before the test function finishes.
        """
        fig = pyplot.figure(FigureClass=cls.FIGURE_CLASS)
        yield fig
        cls.save_and_close(fig)


class AxesTestBase(_Base):
    """Base class for tests of `Axes`."""

    AXES_CLASS: type[Axes] = Axes

    @pytest.fixture
    @classmethod
    def ax(cls):
        """Yield a new instance of `.AXES_CLASS`.

        This fixture checks that the figure can be rendered as a PNG (in memory)
        before the test function finishes.
        """
        fig = pyplot.figure(FigureClass=getattr(cls, "FIGURE_CLASS", Plot))
        yield fig.add_subplot(projection=cls.AXES_CLASS.name)
        cls.save_and_close(fig)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Interface to the LIGO Channel Information System."""

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    TypedDict,
)

import numpy

from .. import (
    Channel,
    ChannelList,
)

if TYPE_CHECKING:
    from typing import NotRequired

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

# -- CIS definitions -----------------

CIS_API_URL = "https://cis.ligo.org/api/channel"
CIS_DATA_TYPE = {
    1: numpy.int16,
    2: numpy.int32,
    3: numpy.int64,
    4: numpy.float32,
    5: numpy.float64,
    6: numpy.complex64,
    7: numpy.uint32,
}


class CisChannelResponse(TypedDict):
    """Response from the CIS for one channel."""

    datarate: float
    datatype: int
    displayurl: str
    name: str
    source: str
    units: str


class CisResponse(TypedDict):
    """Response from a CIS API query."""

    # list of channels
    results: list[CisChannelResponse]
    # pager
    next: NotRequired[str]


# -- I/O routines --------------------

def query(
    name: str | Channel,
    *,
    kerberos: bool | None = None,
    **kwargs,
) -> ChannelList:
    """Query the Channel Information System for details on a channel.

    Parameters
    ----------
    name : `~gwpy.detector.Channel`, `str`
        Name of the channel of interest.

    kerberos : `bool`, `None`, optional
        Use an existing Kerberos ticket as the authentication credential.
        Default behaviour (`kerberos=None`) is to check for credentials
        and request username and passowrd (interactively) if none are foudn.

    kwargs
        Other keyword arguments are passed directly to :func:`ciecplib.get`.

    Returns
    -------
    channel : `~gwpy.detector.Channel`
        Channel with all details as acquired from the CIS
    """
    out = ChannelList()
    url: str | None = f"{CIS_API_URL}/?q={name}"
    while url:
        reply = _get(url, kerberos=kerberos, **kwargs)

        # list result
        if not isinstance(reply, dict):
            out.extend(map(_parse_json, reply))
            break

        # (paged) result
        out.extend(map(_parse_json, reply["results"]))
        url = reply.get("next", None)

    out.sort(key=lambda c: c.name)
    return out


def _get(
    url: str,
    *,
    kerberos: bool | None = None,
    idp: str = "login.ligo.org",
    **kwargs,
) -> CisResponse | list[CisChannelResponse]:
    """Perform a GET query against the CIS."""
    from ciecplib import get as get_cis
    response = get_cis(url, endpoint=idp, kerberos=kerberos, **kwargs)
    response.raise_for_status()
    return response.json()


def _parse_json(
    data: CisChannelResponse,
) -> Channel:
    """Parse the input data dict into a `Channel`.

    Parameters
    ----------
    data : `dict`
        Input data from CIS json query.

    Returns
    -------
    c : `Channel`
        A `Channel` built from the data.
    """
    sample_rate = data["datarate"]
    unit = data["units"]
    dtype = CIS_DATA_TYPE[data["datatype"]]
    model = data["source"]
    url = data["displayurl"]
    return Channel(
        data["name"],
        sample_rate=sample_rate,
        unit=unit,
        dtype=dtype,
        model=model,
        url=url,
    )



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""tconvert: a utility to convert to and from GPS times.

This method is inspired by the original tconvert utility, written by
Peter Shawhan.
"""

from __future__ import annotations

import datetime
from decimal import Decimal
from numbers import Number
from typing import (
    TYPE_CHECKING,
    SupportsFloat,
    cast,
    overload,
)

from astropy.time import Time
from astropy.units import Quantity
from dateparser import parse as dateparser_parse

from . import LIGOTimeGPS
from ._ligotimegps import LIGOTimeGPSLike

if TYPE_CHECKING:
    from collections.abc import Callable

SupportsToGps = LIGOTimeGPSLike | SupportsFloat | datetime.date | Time | str

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__all__ = [
    "from_gps",
    "tconvert",
    "to_gps",
]


@overload
def tconvert(
    gpsordate: SupportsFloat,
) -> datetime.datetime:
    ...

@overload
def tconvert(
    gpsordate: datetime.date | Time | str,
) -> LIGOTimeGPS:
    ...

def tconvert(
    gpsordate: SupportsToGps = "now",
) -> datetime.date | LIGOTimeGPS:
    """Convert GPS times to ISO-format date-times and vice-versa.

    .. warning::

       This method cannot convert exact leap seconds to
       `datetime.datetime`, that object doesn't support it,
       so you should consider using `astropy.time.Time` directly.

    Parameters
    ----------
    gpsordate : `float`, `astropy.time.Time`, `datetime.datetime`, ...
        Input gps or date to convert, many input types are supported.

    Returns
    -------
    date : `datetime.datetime` or `LIGOTimeGPS`
        Converted gps or date.

    Notes
    -----
    - If the input object is a `float` or `LIGOTimeGPS`, it will get
      converted from GPS format into a `datetime.datetime`, otherwise
      the input will be converted into `LIGOTimeGPS`.

    - All `~datetime.datetime` objects returned by this function are 'aware'
      in the UTC timezone.
      For more details on timezone handling in `~datetime.datetime` objects,
      see :ref:`python:datetime-naive-aware`.

    Examples
    --------
    Integers and floats are automatically converted from GPS to
    `datetime.datetime`:

    >>> from gwpy.time import tconvert
    >>> tconvert(0)
    datetime.datetime(1980, 1, 6, 0, 0, tzinfo=datetime.timezone.utc)
    >>> tconvert(1126259462.3910)
    datetime.datetime(2015, 9, 14, 9, 50, 45, 391000, tzinfo=datetime.timezone.utc)

    while strings are automatically converted to `~gwpy.time.LIGOTimeGPS`:

    >>> to_gps("Sep 14 2015 09:50:45.391")
    LIGOTimeGPS(1126259462, 391000000)

    Additionally, a few special-case words as supported, which all return
    `~gwpy.time.LIGOTimeGPS`:

    >>> tconvert("now")
    >>> tconvert("today")
    >>> tconvert("tomorrow")
    >>> tconvert("yesterday")
    """
    # convert from GPS into datetime
    try:
        # if we can 'float' it, then its probably a GPS time
        float(gpsordate)  # type: ignore[arg-type]
    except (
        TypeError,
        ValueError,
    ):
        return to_gps(gpsordate)
    return from_gps(cast("SupportsFloat", gpsordate))


def to_gps(
    t: SupportsToGps,
    *,
    tzinfo: datetime.tzinfo = datetime.UTC,
) -> LIGOTimeGPS:
    """Convert any input date/time into a `LIGOTimeGPS`.

    Any input object that can be cast as a `~astropy.time.Time`
    (with `str` going through the `datetime.datetime`) are acceptable.

    Parameters
    ----------
    t : `float`, `~datetime.datetime`, `~astropy.time.Time`, `str`
        The input time, any object that can be converted into a
        `LIGOTimeGPS`, `~astropy.time.Time`, or `~datetime.datetime`,
        is acceptable.

    tzinfo : `datetime.tzinfo`,
        Timezone information to attach to `tuple` inputs that
        become `datetime.datetime` objects.
        Defaults to `datetime.UTC`.

    Returns
    -------
    gps : `LIGOTimeGPS`
        The number of GPS seconds (non-integer) since the start of the
        epoch (January 6 1980).

    Raises
    ------
    TypeError
        If a `str` input cannot be parsed as a `datetime.datetime`.
    ValueError
        If the input cannot be cast as a `~astropy.time.Time` or
        `LIGOTimeGPS`.

    Examples
    --------
    >>> to_gps("Jan 1 2017")
    LIGOTimeGPS(1167264018, 0)
    >>> to_gps("Sep 14 2015 09:50:45.391")
    LIGOTimeGPS(1126259462, 391000000)

    >>> import datetime
    >>> to_gps(datetime.datetime(2017, 1, 1))
    LIGOTimeGPS(1167264018, 0)

    >>> from astropy.time import Time
    >>> to_gps(Time(57754, format="mjd"))
    LIGOTimeGPS(1167264018, 0)
    """
    if isinstance(t, LIGOTimeGPS):
        return t
    if isinstance(t, LIGOTimeGPSLike):
        return LIGOTimeGPS(
            t.gpsSeconds,
            t.gpsNanoSeconds,
        )

    # -- convert input to Time, or something we can pass to LIGOTimeGPS

    if isinstance(t, str):
        try:  # if str represents a number, leave it for LIGOTimeGPS to handle
            float(t)
        except ValueError:  # str -> datetime.date
            t = _str_to_datetime(t)

    # tuple -> datetime.date
    if isinstance(t, tuple | list):
        t = datetime.datetime(*t, tzinfo=tzinfo)  # ty: ignore

    # datetime.datetime -> Time
    if isinstance(t, datetime.date):
        t = _datetime_to_time(t)

    # Quantity -> float
    if isinstance(t, Quantity):
        t = t.to("second").value

    # Number/Decimal -> str
    if isinstance(t, Decimal | Number):
        t = str(t)

    # -- convert to LIGOTimeGPS

    if isinstance(t, Time):
        return _time_to_gps(t)
    try:
        return LIGOTimeGPS(t)  # type: ignore[arg-type]
    except (TypeError, ValueError):
        return LIGOTimeGPS(float(t))


def from_gps(
    gps: SupportsFloat,
) -> datetime.datetime:
    """Convert a GPS time into a `datetime.datetime`.

    .. warning::

       This method cannot convert exact leap seconds to
       `datetime.datetime`, that object doesn't support it,
       so you should consider using `astropy.time.Time` directly.

    Parameters
    ----------
    gps : `LIGOTimeGPS`, `int`, `float`
        GPS time to convert.

    Returns
    -------
    datetime : `datetime.datetime`
        UTC-aware `~datetime.datetime` representation of the input GPS time.

    Notes
    -----
    - All `~datetime.datetime` objects returned by this function are 'aware'
      in the UTC timezone.
      For more details on timezone handling in `~datetime.datetime` objects,
      see :ref:`python:datetime-naive-aware`.

    Examples
    --------
    >>> from_gps(1167264018)
    datetime.datetime(2017, 1, 1, 0, 0)
    >>> from_gps(1126259462.3910)
    datetime.datetime(2015, 9, 14, 9, 50, 45, 391000)
    """
    try:
        ltgps = LIGOTimeGPS(gps)  # type: ignore[call-arg,call-overload]
    except (
        ValueError,
        TypeError,
        RuntimeError,
    ):
        ltgps = LIGOTimeGPS(float(gps))
    sec, nano = ltgps.gpsSeconds, ltgps.gpsNanoSeconds
    try:
        date = Time(sec, format="gps", scale="utc").datetime
    except ValueError as exc:
        if "within a leap second" in str(exc):
            exc.args = (
                "cannot represent leap second using datetime.datetime, "
                "consider using "
                f'astropy.time.Time({gps}, format="gps", scale="utc") '
                "directly",
            )
        raise
    return (
        date.replace(tzinfo=datetime.UTC)
        + datetime.timedelta(microseconds=nano * 1e-3)
    )


# -- utilities ----------------------------------------------------------------
# special case strings

def _now() -> datetime.datetime:
    now = datetime.datetime.now(datetime.UTC)
    return now.replace(microsecond=0)


def _today() -> datetime.date:
    return _now().date()


def _today_delta(**delta) -> datetime.date:
    return _today() + datetime.timedelta(**delta)


def _tomorrow() -> datetime.date:
    return _today_delta(days=1)


def _yesterday() -> datetime.date:
    return _today_delta(days=-1)


DATE_STRINGS: dict[str, Callable] = {
    "now": _now,
    "today": _today,
    "tomorrow": _tomorrow,
    "yesterday": _yesterday,
}


def _str_to_datetime(
    datestr: str,
) -> datetime.datetime:
    """Convert `str` to `datetime.datetime`."""
    # try known string
    try:
        return DATE_STRINGS[str(datestr).lower()]()
    except KeyError:  # any other string
        pass

    result = dateparser_parse(
        datestr,
        settings={
            "TIMEZONE": "UTC",
            "RETURN_AS_TIMEZONE_AWARE": True,
            "TO_TIMEZONE": "UTC",
            "PREFER_DATES_FROM": "current_period",
        },
    )
    if result is None:
        msg = f"failed to parse '{datestr}' as datetime"
        raise ValueError(msg)
    return result


def _datetime_to_time(
    dtm: datetime.date,
) -> Time:
    """Convert `datetime.date` to `astropy.time.Time`."""
    # astropy.time.Time requires datetime.datetime
    if not isinstance(dtm, datetime.datetime):
        dtm = datetime.datetime.combine(dtm, datetime.time.min)
    return Time(dtm, scale="utc")


def _time_to_gps(
    time: Time,
) -> LIGOTimeGPS:
    """Convert `astropy.time.Time` into `LIGOTimeGPS`.

    This method uses `datetime.datetime` underneath, which restricts
    to microsecond precision by design. This should probably be fixed...

    Parameters
    ----------
    time : `~astropy.time.Time`
        Time object to convert.

    Returns
    -------
    gps : `LIGOTimeGPS`
        Nano-second precision `LIGOTimeGPS` time.
    """
    time = time.utc
    date = time.datetime
    micro = date.microsecond if isinstance(date, datetime.datetime) else 0
    return LIGOTimeGPS(int(time.gps), int(micro * 1e3))



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.cli`."""

from __future__ import annotations

import contextlib
import warnings
from argparse import ArgumentParser
from typing import (
    TYPE_CHECKING,
    Generic,
    TypeVar,
    cast,
)

import pytest
from matplotlib import pyplot
from numpy.random import default_rng

from ...plot import Plot
from ...testing import utils
from ...timeseries import TimeSeries
from .. import (
    SpectrumProduct,
    TransferFunctionProduct,
    cliproduct,
)

if TYPE_CHECKING:
    from argparse import Namespace
    from pathlib import Path
    from typing import ClassVar

    NamespaceType = TypeVar("NamespaceType", bound=Namespace)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

CliProductType = TypeVar("CliProductType", bound=cliproduct.CliProduct)
ImageProductType = TypeVar("ImageProductType", bound=cliproduct.ImageProduct)
TimeDomainProductType = TypeVar(
    "TimeDomainProductType",
    bound=cliproduct.TimeDomainProduct,
)
FrequencyDomainProductType = TypeVar(
    "FrequencyDomainProductType",
    bound=cliproduct.FrequencyDomainProduct,
)

RNG0 = default_rng(0)
RNG1 = default_rng(1)

# set data for nds2_connection test fixture
NDS2_CONNECTION_FIXTURE_DATA = [
    TimeSeries(
        RNG0.random(10240),
        t0=0,
        sample_rate=1024,
        name="X1:TEST-CHANNEL",
    ),
    TimeSeries(
        RNG1.random(10240),
        t0=0,
        sample_rate=1024,
        name="Y1:TEST-CHANNEL",
    ),
]


# -- utilities -----------------------

def update_namespace(args: NamespaceType, **params) -> NamespaceType:
    """Update the `argparse.Namespace` with new parameters."""
    for key, value in params.items():
        setattr(args, key, value)
    return args


# -- function tests ------------------

def test_to_float():
    """Test `to_float`."""
    to_s = cliproduct.to_float("s")
    s = to_s(3)
    assert isinstance(s, float)
    assert s == 3.
    assert to_s("3ms") == 0.004


# -- class tests ---------------------

class _TestCliProduct(Generic[CliProductType]):
    """Test the `CliProduct` class."""

    ACTION: ClassVar[str | None] = None
    TEST_CLASS: ClassVar[type[cliproduct.CliProduct]] = cliproduct.CliProduct  # type: ignore[type-abstract]
    TEST_ARGS: ClassVar[list[str]] = [
        "--chan", "X1:TEST-CHANNEL",
        "--start", "0",
        "--nds2-server", "nds.test.gwpy",
        "--dpi", "100",
        "--geometry", "640x480",
    ]
    NDS2_CONNECTION_FIXTURE_DATA = NDS2_CONNECTION_FIXTURE_DATA

    # -- fixtures --------------------

    @pytest.fixture
    @classmethod
    def args(cls) -> Namespace:
        """Create and parser arguments for a given `CliProduct`.

        Returns the `argparse.Namespace`
        """
        parser = ArgumentParser()
        parser.add_argument("--verbose", action="count", default=1)
        parser.add_argument("--silent", action="store_true")
        cls.TEST_CLASS.init_cli(parser)
        return parser.parse_args([str(x) for x in cls.TEST_ARGS])

    @pytest.fixture
    @classmethod
    def prod(cls, args):
        """Return a `CliProduct`."""
        prod = cls.TEST_CLASS(args)
        yield prod
        with contextlib.suppress(RuntimeError):
            prod.plot.close()

    @staticmethod
    def _prod_add_data(prod: CliProductType) -> CliProductType:
        # we need this method separate, rather than in dataprod, so that
        # we can have classes override the dataprod fixture with extra
        # stuff properly
        dur = prod.duration
        fs = 512

        i = 0
        for start in prod.start_list:
            for chan in prod.chan_list:
                rng = default_rng(i)
                ts = TimeSeries(
                    rng.random(int(fs*dur)),
                    t0=start,
                    sample_rate=512,
                    name=chan,
                )
                prod.timeseries.append(ts)
                i += 1
        return prod

    @pytest.fixture
    @classmethod
    def dataprod(cls, prod):
        """Return a `CliProduct` with data."""
        return cls._prod_add_data(prod)

    @staticmethod
    def _plotprod_init(prod: CliProductType) -> CliProductType:
        prod.plot = cast("Plot", pyplot.figure(FigureClass=Plot))
        prod.plot.gca()  # Initialise some axes
        return prod

    @pytest.fixture
    @classmethod
    def plotprod(cls, dataprod):
        """Return a `CliProduct` with data and a plot."""
        return cls._plotprod_init(dataprod)

    # -- tests -----------------------

    def test_init(self, args: Namespace):
        """Test `CliProduct` initialization."""
        channels = [
            self.TEST_ARGS[i+1]
            for i, arg in enumerate(self.TEST_ARGS)
            if arg == "--chan"
        ]
        prod = self.TEST_CLASS(args)
        with pytest.raises(RuntimeError):
            assert prod.plot is None
        assert prod.plot_num == 0
        assert prod.start_list == [0]
        assert prod.duration == 10.
        assert prod.chan_list == channels
        assert prod.n_datasets == len(prod.start_list) * len(prod.chan_list)
        assert prod.timeseries == []
        assert prod.dpi == 100.
        assert prod.width == 640
        assert prod.height == 480
        assert prod.figsize == (6.4, 4.8)

    def test_action(self, prod: CliProductType):
        """Test `CliProduct.action`."""
        if self.ACTION is None:
            raise NotImplementedError
        assert prod.action is self.ACTION

    @pytest.mark.parametrize("level", [
        pytest.param(1, id="info"),
        pytest.param(2, id="debug"),
    ])
    def test_log(
        self,
        prod: CliProductType,
        level: int,
        caplog: pytest.LogCaptureFixture,
    ):
        """Test `CliProduct.log`."""
        caplog.set_level("INFO", logger=prod.logger.name)
        with pytest.warns(
            DeprecationWarning,
            match="`CliProduct.log` is deprecated",
        ):
            prod.log(level, "Test")
        assert ("Test" in caplog.text) is (level <= 1)

    @pytest.mark.requires("nds2")
    @pytest.mark.usefixtures("nds2_connection")
    def test_get_data(self, prod: CliProductType):
        """Test `CliProduct.get_data`."""
        prod.get_data()

        utils.assert_quantity_sub_equal(
            prod.timeseries[0],
            NDS2_CONNECTION_FIXTURE_DATA[0],
            exclude=("channel",),
        )

    @pytest.mark.parametrize(("ftype", "filt"), [
        pytest.param("highpass", {"highpass": 10}, id="highpass"),
        pytest.param("lowpass", {"lowpass": 100}, id="lowpass"),
        pytest.param("bandpass", {"highpass": 10, "lowpass": 100}, id="bandpass"),
    ])
    def test_filter_timeseries(
        self,
        dataprod: CliProductType,
        ftype: str,
        filt: dict[str, float],
    ):
        """Test `CliProduct._filter_timeseries`."""
        ts = dataprod.timeseries[0]
        if ftype == "bandpass":
            result = ts.bandpass(filt["highpass"], filt["lowpass"])
        else:
            result = cast("TimeSeries", getattr(ts, ftype)(filt[ftype]))

        fts = dataprod._filter_timeseries(ts, **filt)  # type: ignore[arg-type]
        utils.assert_quantity_sub_equal(fts, result)

    @pytest.mark.parametrize(("params", "title"), [
        pytest.param(
            {"highpass": 100},
            "high pass (100.0)",
            id="highpass",
        ),
        pytest.param(
            {"lowpass": 100},
            "low pass (100.0)",
            id="lowpass",
        ),
        pytest.param(
            {"highpass": 100, "lowpass": 200},
            "band pass (100.0-200.)",
            id="highlowband",
        ),
        pytest.param(
            {"highpass": 100, "notch": [60]},
            "high pass (100.0), notch(60.0)",
            id="highnotch",
        ),
    ])
    def test_get_title(
        self,
        prod: CliProductType,
        params: dict[str, float],
        title: str,
    ):
        """Test `CliProduct.get_title`."""
        update_namespace(prod.args, **params)  # update parameters
        assert prod.get_title() == title

    def test_get_suptitle(self, prod: CliProductType):
        """Test `CliProduct.get_suptitle`."""
        assert prod.get_suptitle() == prod.chan_list[0]

    @pytest.mark.parametrize("params", [
        pytest.param({}, id="default"),
        pytest.param(
            {
                "xscale": "linear",
                "xmin": 0,
                "xmax": 5,
                "xlabel": "X-label",
                "yscale": "log",
                "ymin": 0,
                "ymax": 50,
                "ylabel": "Y-label",
            },
            id="custom",
        ),
    ])
    def test_set_plot_properties(
        self,
        plotprod: CliProductType,
        params: dict[str, float],
    ):
        """Test `CliProduct.set_plot_properties`."""
        args = plotprod.args
        update_namespace(args, **params)  # update parameters

        if isinstance(plotprod, TransferFunctionProduct):
            data = plotprod.tfs
        elif isinstance(plotprod, SpectrumProduct):
            data = plotprod.spectra
        else:
            data = plotprod.timeseries
        xmin = min(series.xspan[0] for series in data)
        xmax = max(series.xspan[1] for series in data)

        # ignore warnings from matplotlib about having no labels
        # (because we have cut some corners in preparing this test)
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message="No artists with labels found to put in legend",
                category=UserWarning,
            )
            plotprod.set_plot_properties()

        ax = plotprod.ax

        ymin, ymax = ax.get_ylim()

        assert ax.get_xscale() == params.get("xscale", args.xscale)
        assert ax.get_xlim() == (params.get("xmin", xmin),
                                 params.get("xmax", xmax))
        assert ax.get_xlabel() == params.get("xlabel", plotprod.get_xlabel())

        assert ax.get_yscale() == params.get("yscale", args.yscale or "linear")
        assert ax.get_ylim() == (params.get("ymin", ymin),
                                 params.get("ymax", ymax))
        assert ax.get_ylabel() == params.get("ylabel", plotprod.get_ylabel())

    @pytest.mark.requires("nds2")
    @pytest.mark.usefixtures("nds2_connection")
    def test_run(
        self,
        tmp_path: Path,
        prod: CliProductType,
    ):
        """Test `CliProduct.run`."""
        tmp = tmp_path / "plot.png"
        prod.args.out = str(tmp)
        prod.run()
        assert tmp.is_file()
        assert prod.plot_num == 1
        assert not prod.has_more_plots()


class _TestImageProduct(_TestCliProduct[ImageProductType], Generic[ImageProductType]):
    """Test the `ImageProduct` class."""

    TEST_CLASS: ClassVar[type[cliproduct.ImageProduct]] = cliproduct.ImageProduct  # type: ignore[type-abstract]

    def test_extra_plot_options(self, args: Namespace):
        """Test `ImageProduct.extra_plot_options`."""
        for key in ("nocolorbar", "cmap", "imin", "imax"):
            assert hasattr(args, key)

    @pytest.mark.parametrize("visible", [False, True])
    def test_set_plot_properties(
        self,
        plotprod: ImageProductType,
        visible: bool,
    ):
        """Test `ImageProduct.set_plot_properties`."""
        update_namespace(plotprod.args, nocolorbar=not visible)
        plotprod.set_plot_properties()
        if visible:
            label = plotprod.get_color_label()
            coloraxes = plotprod.ax.child_axes[0]
            assert coloraxes.get_ylabel() == label
        else:
            # check our axes doesn't have any children
            assert not plotprod.ax.child_axes


class _TestFFTMixin:
    """Test the `FFTMixin` class."""


class _TestTimeDomainProduct(
    _TestCliProduct[TimeDomainProductType],
    Generic[TimeDomainProductType],
):
    """Test the `TimeDomainProduct` class."""

    TEST_CLASS: ClassVar[type[cliproduct.TimeDomainProduct]] = (
        cliproduct.TimeDomainProduct  # type: ignore[type-abstract]
    )


class _TestFrequencyDomainProduct(
    _TestCliProduct[FrequencyDomainProductType],
    Generic[FrequencyDomainProductType],
):
    """Test the `FrequencyDomainProduct` class."""

    TEST_CLASS: ClassVar[type[cliproduct.FrequencyDomainProduct]] = (
        cliproduct.FrequencyDomainProduct  # type: ignore[type-abstract]
    )



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwpy.io.root`."""

import pytest

from .. import root as io_root

uproot = pytest.importorskip("uproot")


@pytest.fixture
def rootfile(tmp_path):
    """Create an empty ROOT file using `uproot.create`."""
    path = tmp_path / "test.root"
    with uproot.create(path):
        pass
    return path


def test_identify_root_fileobj(rootfile):
    """Test that `gwpy.io.root.identify_root` works with file objects."""
    with open(rootfile, "rb") as file:
        assert io_root.identify_root(
            "read",
            None,
            file,
        )


def test_identify_root_filepath(rootfile):
    """Test that `gwpy.io.root.identify_root` works with file paths."""
    assert io_root.identify_root(
        "read",
        str(rootfile),
        None,
    )


def test_identify_root_uproot(rootfile):
    """Test that `gwpy.io.root.identify_root` works with `uproot` objects."""
    with uproot.open(rootfile) as rootf:
        assert io_root.identify_root(
            "read",
            None,
            None,
            rootf,
        )



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Input/Output routines for the TimeSeries and its sub-classes."""

from . import (  # pylint: disable=unused-import
    ascii,
    gwdatafind,
    gwf,
    hdf5,
    losc,
    nds2,
    wav,
)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read events from ROOT trees into Tables."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ...io.root import identify_root
from ...io.utils import file_path
from .. import (
    EventTable,
    Table,
)
from .utils import read_with_columns_and_where

if TYPE_CHECKING:
    from pathlib import Path
    from typing import IO

    import uproot

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def get_treename(
    rootdir: uproot.ReadOnlyDirectory,
) -> str:
    """Return the name of the only tree in this root directory.

    Parameters
    ----------
    rootdir : `uproot.reading.ReadOnlyDirectory`
        The ROOT object to inspect.

    Raises
    ------
    ValueError
        If multiple trees are found.
    """
    source = rootdir.file_path

    if not rootdir:  # nothing to read?
        msg = "No trees found in '{source}'"
        raise ValueError(msg)

    # find one and only one tree
    try:
        tree, = rootdir.iterkeys(
            recursive=True,
            cycle=False,
        )
    except ValueError as exc:
        msg = (
            f"Multiple trees found in {source}, please select one via the "
            "`treename` keyword argument, e.g. `treename='events'`. "
            "Available trees are: '{', '.join(names)}'."
        )
        raise ValueError(msg) from exc
    return tree


@read_with_columns_and_where
def table_from_root(
    source: str | Path | IO,
    treename: str | None = None,
    **kwargs,
) -> Table:
    """Read a Table from a ROOT file.

    Requires: :doc:`uproot <uproot:index>`

    Parameters
    ----------
    source : `str`, `pathlib.Path`
        The file path or object from which to read. See `uproot.open`
        for details on acceptable inputs.

    treename : `str`, optional
        The name of the ``TTree`` to read.
        Required unless ``source`` contains exactly one ``TTree``.

    columns : `list` of `str`, optional
        List of column names to read.

    where : `str`, `list` of `str`, optional
        One or more column filters with which to downselect the
        returned table rows as they as read, e.g. ``'snr > 5'``,
        similar to a SQL ``WHERE`` statement.
        Multiple conditions should be connected by ' && ' or ' and ',
        or given as a `list`, e.g. ``'snr > 5 && frequency < 1000'`` or
        ``['snr > 5', 'frequency < 1000']``.

    kwargs
        All other keyword arguments are either passed to
        `uproot.open` or to the `~astropy.table.Table` constructor.

    Raises
    ------
    ValueError
        If ``treename=None`` is given and multiple trees exist in
        the ``source``.

    KeyError
        If ``treename`` is given but no tree is found with that name.

    See Also
    --------
    uproot.open
        For details of how ROOT files are parsed and what keyword
        arguments should be supported.

    astropy.table.Table
        For details of the keyword arguments supported when creating tables.
    """
    import uproot

    # handle uproot.open keywords
    createkw = {
        # valid for uproot 5.5.1
        k: kwargs.pop(k) for k in (
            "object_cache",
            "array_cache",
            "custom_classes",
            "decompression_executor",
            "interpretation_executor",
            "handler",
            "timeout",
            "max_num_elements"
            "num_workers"
            "use_threads"
            "num_fallback_workers"
            "begin_chunk_size"
            "minimal_ttree_metadata",
        ) if k in kwargs
    }

    path = file_path(source)
    with uproot.open(path, **createkw) as rootdir:
        # find tree name
        if treename is None:
            treename = get_treename(rootdir)

        # read branches from tree
        return Table(rootdir[treename].arrays(library="np"), **kwargs)


def table_to_root(  # noqa: D417
    table: Table,
    filename: str | Path | IO,
    treename: str = "tree",
    *,
    overwrite: bool = False,
    append: bool =False,
    **kwargs,
) -> None:
    """Write a Table to a ROOT file.

    Requires: :doc:`uproot <uproot:index>`

    Parameters
    ----------
    filename : `str`, `pathlib.Path`
        Filename or object to write to.

    treename : `str`
        Name of the ROOT ``TTree`` to create for this table.

    overwrite : `bool`
        If `True` over-write an existing file of the same name.
        Default is `False`.

    append : `bool`
        If `True` append a new ``TTree`` to an existing file.
        Default is `False`.

    kwargs
        All other keyword arguments are passed to the relevant
        `uproot` functions for creating/updating a file and
        writing a ``TTree``.

    Raises
    ------
    FileExistsError
        If ``overwrite=False, append=False`` is given and the
        target filename already exists.

    OSError
        If ``append=True`` is given and the target filename does not
        already exist.

    See Also
    --------
    uproot.create
        For details of how new files are created and what keyword
        arguments should be supported.
        This is called if ``overwrite=False, append=False`` is given.

    uproot.update
        For details of how existing files are updated and what keyword
        arguments should be supported.
        This is called if ``append=True`` is given.

    uproot.recreate
        For details of how existing files are over-written and what keyword
        arguments should be supported.
        This is called if ``overwrite=True, append=False`` is given.

    uproot.writing.writable.WritableDirectory.mktree
        For deatils of how ``TTree`` objects are created and what keyword
        arguments should be supported.
    """
    import uproot

    # handle file creation/update options
    createkw = {
        k: kwargs.pop(k) for k in (
            "initial_directory_bytes",
            "initial_streamers_bytes",
            "uuid_function",
            "compression",
        ) if k in kwargs
    }
    if overwrite:
        create_func = uproot.recreate
    elif append:
        create_func = uproot.update
    else:
        create_func = uproot.create

    # create file
    with create_func(filename, **createkw) as outf:
        # create the tree
        tree = outf.mktree(
            treename,
            dict(table.dtype.descr),
            **kwargs,
        )
        # add data to it
        tree.extend(dict(table.columns))


# register I/O
for klass in (Table, EventTable):
    klass.read.registry.register_identifier("root", klass, identify_root)
    klass.read.registry.register_reader("root", klass, table_from_root)
    klass.write.registry.register_writer("root", klass, table_to_root)



# Copyright (c) 2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""
.. sectionauthor:: Duncan Macleod <duncan.macleod@ligo.org>
.. currentmodule:: gwpy.timeseries

Calculating the SNR associated with an astrophysical signal model
#################################################################

The example :ref:`sphx_glr_examples_signal_gw150914.py` showed us we can visually
extract a signal from the noise using basic signal-processing techniques.

However, an actual astrophysical search algorithm detects signals by
calculating the signal-to-noise ratio (SNR) of data for each in a large bank
of signal models, known as templates.

Using :doc:`PyCBC <pycbc:index>` (the actual search code), we can do that.
"""

# %%
# Data access and conditioning
# ----------------------------
# First we fetch some of the public data from |GWOSC|_:

from gwpy.timeseries import TimeSeries
data = TimeSeries.get("H1", 1126259446, 1126259478)

# %%
# and condition it by applying a highpass filter at 15 Hz

high = data.highpass(15)

# %%
# This is important to remove noise at lower frequencies that isn't
# accurately calibrated, and swamps smaller noises at higher frequencies.
#
# For this example, we want to calculate the SNR over a 4 second segment, so
# we calculate a Power Spectral Density with a 4 second FFT length (using all
# of the data), then :meth:`~TimeSeries.crop` the data:

psd = high.psd(4, 2)
zoom = high.crop(1126259460, 1126259464)

# %%
# Generating a signal model
# -------------------------
# In order to calculate signal-to-noise ratio, we need a signal model
# against which to compare our data.
# For this we import :func:`pycbc.waveform.get_fd_waveform
# <pycbc.waveform.waveform.get_fd_waveform>` and generate a template as a
# `pycbc.types.FrequencySeries <pycbc.types.frequencyseries.FrequencySeries>`:

from pycbc.waveform import get_fd_waveform
hp, _ = get_fd_waveform(
    approximant="IMRPhenomD",
    mass1=40,
    mass2=32,
    f_lower=20,
    f_final=2048,
    delta_f=psd.df.value,
)

# %%
# Calculating SNR
# ---------------
# At this point we are ready to calculate the SNR, so we import the
# :func:`pycbc.filter.matched_filter
# <pycbc.filter.matchedfilter.matched_filter>` method, and pass it
# our template, the data, and the PSD:

from pycbc.filter import matched_filter
snr = matched_filter(
    hp,
    zoom.to_pycbc(),
    psd=psd.to_pycbc(),
    low_frequency_cutoff=15,
)
snrts = TimeSeries.from_pycbc(snr).abs()

# %%
# .. tip::
#
#    Here we have used the :meth:`~TimeSeries.to_pycbc` methods of the
#    `~gwpy.timeseries.TimeSeries` and `~gwpy.frequencyseries.FrequencySeries`
#    objects to convert from GWpy objects to something that PyCBC functions
#    can understand, and then used the :meth:`~TimeSeries.from_pycbc` method
#    to convert back to a GWpy object.
#
# Visualisation
# -------------
# We can plot the SNR `TimeSeries` around the region of interest:

plot = snrts.plot()
ax = plot.gca()
ax.set_xlim(1126259461, 1126259463)
ax.set_epoch(1126259462.427)
ax.set_ylabel("Signal-to-noise ratio (SNR)")
ax.set_title("LIGO-Hanford signal-correlation for GW150914")
plot.show()

# %%
# We can clearly see a large spike (above 17!) at the time of the
# |GW150914|_ signal!
# This is, in principle, how the full, blind, CBC search is performed, using
# all of the available data, and a bank of tens of thousand of signal models.



# Copyright (c) 2015-2020 Joseph Areeda
#               2020-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Spectrum plots."""

from __future__ import annotations

import logging
import warnings
from typing import TYPE_CHECKING

from astropy.time import Time

from ..plot import Plot
from ..plot.tex import label_to_latex
from ..utils import unique
from .cliproduct import FrequencyDomainProduct

if TYPE_CHECKING:
    from argparse import (
        ArgumentParser,
        Namespace,
        _ArgumentGroup,
    )
    from logging import Logger
    from typing import ClassVar

    from astropy.units import UnitBase

    from ..frequencyseries import FrequencySeries
    from ..plot import Axes

__author__ = "Joseph Areeda <joseph.areeda@ligo.org>"

logger = logging.getLogger(__name__)


class SpectrumProduct(FrequencyDomainProduct):
    """Plot the ASD spectrum of one or more time series."""

    action: ClassVar[str] = "spectrum"

    def __init__(
        self,
        args: Namespace,
        logger: Logger = logger,
    ) -> None:
        """Create a new `Spectrum`."""
        super().__init__(args, logger=logger)
        self.spectra: list[FrequencySeries] = []

    @property
    def units(self) -> list[UnitBase]:
        """The (unique) list of data units for this product."""
        return unique(fs.unit for fs in self.spectra)

    @classmethod
    def arg_xaxis(cls, parser: ArgumentParser) -> _ArgumentGroup:
        """Add an `~argparse._ArgumentGroup` for X-axis options."""
        # use frequency axis on X
        return cls._arg_faxis("x", parser)

    @classmethod
    def arg_yaxis(cls, parser: ArgumentParser) -> _ArgumentGroup:
        """Add an `~argparse._ArgumentGroup` for Y-axis options."""
        # default log Y-axis
        return cls._arg_axis("y", parser, scale="log")

    def _finalize_arguments(self, args: Namespace) -> None:
        """Sanity-check and set defaults for arguments."""
        if args.yscale is None:
            args.yscale = "log"
        super()._finalize_arguments(args)

    def get_ylabel(self) -> str | None:
        """Text for y-axis label."""
        if len(self.units) == 1:
            u = self.units[0].to_string("latex").strip("$")
            return fr"ASD $\left({u}\right)$"
        return "ASD"

    def get_suptitle(self) -> str:
        """Start of default super title, first channel is appended to it."""
        return f"Spectrum: {self.chan_list[0]}"

    def get_title(self) -> str:
        """Return default title for plot."""
        gps = self.start_list[0]
        utc = Time(gps, format="gps", scale="utc").iso
        tstr = f"{utc} | {gps} ({self.duration})"

        fftstr = f"fftlength={self.args.secpfft}, overlap={self.args.overlap}"

        return f"{tstr}, {fftstr}"

    def make_plot(self) -> Plot:
        """Generate the plot from time series and arguments."""
        args = self.args

        fftlength = float(args.secpfft)
        overlap = args.overlap
        method = args.method
        self.logger.debug(
            "Calculating spectrum secpfft: %f, overlap: %f",
            fftlength,
            overlap,
        )
        overlap *= fftlength

        # create plot
        plot = Plot(figsize=self.figsize, dpi=self.dpi)
        ax = plot.gca()

        # handle user specified plot labels
        if self.args.legend:
            nlegargs = len(self.args.legend[0])
        else:
            nlegargs = 0
        if nlegargs > 0 and nlegargs != self.n_datasets:
            warnings.warn(
                "The number of legends specified must match the number of "
                "time series (channels * start times). "
                f"There are {len(self.timeseries)} series "
                f"and {len(self.args.legend)} legends",
                stacklevel=2,
            )
            nlegargs = 0  # don't use  themm

        # determine colour
        colors = self._color_by_ifo()

        for i in range(self.n_datasets):
            series = self.timeseries[i]
            if nlegargs:
                label = self.args.legend[0][i]
            else:
                label = series.name or str(series.channel or "")
                if len(self.start_list) > 1:
                    label += f", {series.t0.value}"

            asd = series.asd(
                fftlength=fftlength,
                overlap=overlap,
                method=method,
            )
            self.spectra.append(asd)

            if self.usetex:
                label = label_to_latex(label)

            # plot
            ax.plot(asd, label=label, color=colors[i])

        if args.xscale == "log" and not args.xmin:
            args.xmin = 1/fftlength

        return plot

    def _scale_axes_from_data(self, ax: Axes) -> None:
        """Restrict data limits for Y-axis based on what you can see."""
        # get tight limits for X-axis
        if self.args.xmin is None:
            self.args.xmin = min(fs.xspan[0] for fs in self.spectra)
        if self.args.xmax is None:
            self.args.xmax = max(fs.xspan[1] for fs in self.spectra)

        # autoscale view for Y-axis
        cropped = [fs.crop(self.args.xmin, self.args.xmax) for fs in self.spectra]
        ymin = min(fs.value.min() for fs in cropped)
        ymax = max(fs.value.max() for fs in cropped)
        ax.yaxis.set_data_interval(ymin, ymax, ignore=True)
        ax.autoscale_view(scalex=False)



# Copyright (c) 2024-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Type annotation tools for GWpy."""

from __future__ import annotations

from collections.abc import Collection
from typing import (
    Self,
    TypeAlias,
)

import numpy
from astropy.units import UnitBase
from astropy.units.typing import QuantityLike
from numpy.typing import (
    ArrayLike,
    DTypeLike,
)

__all__ = [
    "Array1D",
    "ArrayLike1D",
    "UnitLike",
]

#: Type alias for Unit-like types.
UnitLike: TypeAlias = UnitBase | str | None

#: Type alias for 1-dimensional numpy arrays.
Array1D: TypeAlias = numpy.ndarray[tuple[int]]
#: Type alias for 1-dimensional array-like types.
ArrayLike1D = Collection[complex] | Array1D



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read events from Coherent Wave-Burst (cWB)-format ROOT and ASCII files."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING

from astropy.io.ascii import core
from astropy.io.registry import get_reader as astropy_get_reader

from .. import (
    EventTable,
    Table,
)
from .utils import read_with_columns_and_where

if TYPE_CHECKING:
    from pathlib import Path
    from typing import IO

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


# -- ROOT ----------------------------

def table_from_cwb_root(
    source: str | Path | IO,
    treename: str = "waveburst",
    **kwargs,
) -> Table:
    """Read an `EventTable` from a Coherent WaveBurst ROOT file.

    This function just redirects to the format='root' reader with appropriate
    defaults.

    See `EventTable.read.help('root')` for more details.
    """
    return Table.read(
        source,
        format="root",
        treename=treename,
        **kwargs,
    )


def table_to_cwb_root(
    table: Table,
    target: str | Path | IO,
    treename: str = "waveburst",
    **kwargs,
) -> None:
    """Read an `EventTable` from a Coherent WaveBurst ROOT file.

    This function just redirects to the format='root' reader with appropriate
    defaults.

    See `EventTable.write.help('root')` for more details.
    """
    return table.write(
        target,
        treename=treename,
        **kwargs,
    )


EventTable.read.registry.register_reader(
    "root.cwb",
    EventTable,
    table_from_cwb_root,
)
EventTable.write.registry.register_writer(
    "root.cwb",
    EventTable,
    table_to_cwb_root,
)


# -- ASCII ---------------------------

class CwbHeader(core.BaseHeader):
    """Parser for cWB ASCII header."""

    def get_cols(self, lines: list[str]) -> None:
        """Initialize Column objects from a multi-line ASCII header.

        Parameters
        ----------
        lines : `list`
            List of table lines
        """
        re_name_def = re.compile(
            r"^\s*#\s+"  # whitespace and comment marker
            r"(?P<colnumber>[0-9]+)\s+-\s+"  # number of column
            r"(?P<colname>(.*))",
        )
        self.names = []
        include_cuts = False
        for line in lines:
            if not line:  # ignore empty lines in header (windows)
                continue
            if not line.startswith("# "):  # end of header lines
                break
            if line.startswith("# -/+"):
                include_cuts = True
            else:
                match = re_name_def.search(line)
                if match:
                    self.names.append(match.group("colname").rstrip())

        if not self.names:
            msg = "No column names found in cWB header"
            raise core.InconsistentTableError(msg)

        if include_cuts:
            self.cols = [  # pylint: disable=attribute-defined-outside-init
                core.Column(name="selection cut 1"),
                core.Column(name="selection cut 2"),
            ]
        else:
            self.cols = []  # pylint: disable=attribute-defined-outside-init
        for name in self.names:
            col = core.Column(name=name)
            self.cols.append(col)

    def write(self, lines: list[str]) -> None:
        """Write column headers to ASCII lines."""
        if "selection cut 1" in self.colnames:
            lines.append("# -/+ - not passed/passed final selection cuts")
        for i, name in enumerate(self.colnames):
            lines.append(f"# {i + 1:.2d} - {name}")


class CwbData(core.BaseData):
    """Parser for cWB ASCII data."""

    comment = "#"


class Cwb(core.BaseReader):
    """Read a Cwb file.

    Parameters
    ----------
    Other than the standard keyword options supported for reading ASCII data,
    the following parameters are supported for "ascii.cwb".

    columns : `list` of `str`, optional
        List of column names to read.

    where : `str`, `list` of `str`, optional
        One or more column filters with which to downselect the
        returned table rows as they as read, e.g. ``'snr > 5'``,
        similar to a SQL ``WHERE`` statement.
        Multiple conditions should be connected by ' && ' or ' and ',
        or given as a `list`, e.g. ``'snr > 5 && frequency < 1000'`` or
        ``['snr > 5', 'frequency < 1000']``.
    """

    _format_name = "cwb"
    _io_registry_can_write = True
    _description = "cWB EVENTS format table"

    header_class = CwbHeader
    data_class = CwbData


# register for EventTable
EventTable.read.registry.register_reader(
   "ascii.cwb",
    EventTable,
    read_with_columns_and_where(astropy_get_reader("ascii.cwb", Table)),
)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Definition of a BodePlot."""

from __future__ import annotations

from math import pi
from typing import (
    TYPE_CHECKING,
    cast,
)

import numpy
from matplotlib.ticker import MaxNLocator

from ..signal import filter_design
from . import Plot

if TYPE_CHECKING:
    from astropy.units.typing import QuantityLike
    from matplotlib.axes import Axes
    from matplotlib.lines import Line2D
    from numpy.typing import (
        ArrayLike,
        NDArray,
    )

    from ..frequencyseries import FrequencySeries
    from ..signal.filter_design import FilterType
    from ..typing import Array1D

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__all__ = [
    "BodePlot",
]


def _to_db(a: ArrayLike, *, power: bool = False) -> NDArray:
    """Convert the input array into decibels.

    Parameters
    ----------
    a : `float`, `numpy.ndarray`
        Value or array of values to convert to decibels.

    power : `bool`, optional
        If `True`, convert power-based quantity to dB (:math:`10 * log_{10}(a)`),
        otherwise (default) convert amplitude (:math:`20 * log_{10}(a)`).

    Returns
    -------
    dB : `float`
        The input value converted to decibels.

    Examples
    --------
    >>> to_db(1000)
    30.0
    """
    db = 10 * numpy.log10(a)
    if power:
        return db
    return 2 * db


class BodePlot(Plot):
    """A `Plot` class for visualising transfer functions.

    Parameters
    ----------
    filters : `~scipy.signal.lti`, `~gwpy.frequencyseries.FrequencySeries`
        Any number of the following:

        - linear time-invariant filters, either
          `~scipy.signal.lti` or `tuple` of the following length and form:
          - 2: (numerator, denominator)
          - 3: (zeros, poles, gain)
          - 4: (A, B, C, D)

        - complex-valued `spectra <gwpy.frequencyseries.FrequencySeries>`
          representing a transfer function

    frequencies : `numpy.ndarray`, `int`, optional
        List of frequencies (in Hertz) at which to plot, or an integer
        specifying the number of frequencies to generate.

    dB : `bool`, optional
        If `True`, display magnitude in decibels, otherwise display
        amplitude.

    analog : `bool`, optional
        If `True`, indicates that the input filters are analogue filters.

    sample_rate : `float`, `~astropy.units.Quantity`, optional
        The sampling rate of a digital filter.
        If ``analog=False`` this option is required.

    unit : `str`, optional
        For analogue ZPK filters, the units in which the zeros and poles are
        specified. Either ``'Hz'`` or ``'rad/s'`` (default).

    normalize_gain : `bool`, optional
        Whether to normalize the gain when converting from Hz to rad/s.

        - `False` (default):
          Multiply zeros/poles by -2π but leave gain unchanged.
          This matches the LIGO GDS **'f' plane** convention
          (``plane='f'`` in ``s2z()``).

        - `True`:
          Normalize gain to preserve frequency response magnitude.
          Gain is scaled by :math:`|∏p_i/∏z_i| · (2π)^{(n_p - n_z)}`.
          Use this when your filter was designed with the transfer
          function :math:`H(f) = k·∏(f-z_i)/∏(f-p_i)` in Hz.
          This matches the LIGO GDS **'n' plane** convention
          (``plane='n'`` in ``s2z()``).

        Only used for analogue filters in Hz (``analog=True, unit="Hz"``).

    kwargs
        All other keyword arguments are passed to
        :meth:`~FrequencySeriesAxes.plot`.

    Returns
    -------
    plot : `BodePlot`
        A new BodePlot with two `Axes` - :attr:`~BodePlot.maxes` and
        :attr:`~BodePlot.paxes` - representing the magnitude and
        phase of the input transfer function(s) respectively.
    """

    def __init__(
        self,
        *filters: FilterType,
        dB: bool = True,  # noqa: N803
        frequencies: Array1D | int | None = None,
        **kwargs,
    ) -> None:
        """Initialise a new `BodePlot`."""
        from ..frequencyseries import FrequencySeries

        # parse plotting arguments
        figargs = {}
        title = kwargs.pop("title", None)
        for key in (
            "figsize",
            "dpi",
            "frameon",
            "subplotpars",
            "tight_layout",
        ):
            if key in kwargs:
                figargs[key] = kwargs.pop(key)

        # generate figure
        super().__init__(**figargs)

        # delete the axes, and create two more
        self.add_subplot(2, 1, 1)
        self.add_subplot(2, 1, 2, sharex=self.maxes)

        # add filters
        for filter_ in filters:
            if isinstance(filter_, FrequencySeries):
                self.add_frequencyseries(filter_, dB=dB, **kwargs)
            else:
                self.add_filter(
                    filter_,
                    frequencies=frequencies,
                    dB=dB,
                    **kwargs,
                )

        # format plots
        if dB:
            self.maxes.set_ylabel("Magnitude [dB]")
            ylim = self.maxes.get_ylim()
            if ylim[1] == 0:
                self.maxes.set_ybound(
                    upper=ylim[1] + (ylim[1] - ylim[0]) * 0.1)
        else:
            self.maxes.set_yscale("log")
            self.maxes.set_ylabel("Amplitude")
        if not kwargs.get("analog", False) or kwargs.get("unit") == "Hz":
            self.paxes.set_xlabel("Frequency [Hz]")
        else:
            self.paxes.set_xlabel("Frequency [rad/s]")
        self.paxes.set_ylabel("Phase [deg]")
        self.maxes.set_xscale("log")
        self.paxes.set_xscale("log")
        self.paxes.yaxis.set_major_locator(
            MaxNLocator(nbins="auto", steps=[4.5, 9]),
        )
        self.paxes.yaxis.set_minor_locator(
            MaxNLocator(nbins=20, steps=[4.5, 9]),
        )
        if title:
            self.maxes.set_title(title)

        # get xlim
        if (
            frequencies is None
            and len(filters) == 1
            and isinstance(filters[0], FrequencySeries)
        ):
            frequencies = filters[0].frequencies.value
        if not isinstance(frequencies, type(None) | int):
            frequencies = numpy.asarray(frequencies)
            frequencies = frequencies[frequencies > 0]
            self.maxes.set_xlim(frequencies.min(), frequencies.max())

    @property
    def maxes(self) -> Axes:
        """`FrequencySeriesAxes` for the Bode magnitude."""
        return self.axes[0]

    @property
    def paxes(self) -> Axes:
        """`FrequencySeriesAxes` for the Bode phase."""
        return self.axes[1]

    def add_filter(
        self,
        filter_: FilterType,
        frequencies: Array1D | int | None = None,
        *,
        dB: bool = True,  # noqa: N803
        analog: bool = False,
        sample_rate: QuantityLike = 1.0,
        unit: str | None = None,
        normalize_gain: bool = False,
        **kwargs,
    ) -> tuple[Line2D, Line2D]:
        """Add a linear time-invariant filter to this BodePlot.

        Parameters
        ----------
        filter_ : `~scipy.signal.lti`, `tuple`
            The filter to plot, either as a `~scipy.signal.lti`, or a
            `tuple` with the following number and meaning of elements

            - 2: (numerator, denominator)
            - 3: (zeros, poles, gain)
            - 4: (A, B, C, D)

        frequencies : `numpy.ndarray`, optional
            List of frequencies (in Hertz) at which to plot.

        dB : `bool`, optional
            If `True`, display magnitude in decibels, otherwise display
            amplitude, default: `True`.

        analog : `bool`, optional
            If `True`, indicates that ``filter_`` is an analogue filter.

        sample_rate : `float`, `~astropy.units.Quantity`, optional
            The sampling rate of a digital filter.
            If ``analog=False`` this option is required.

        unit : `str`, optional
            For analogue ZPK filters, the units in which the zeros and poles are
            specified. Either ``'Hz'`` or ``'rad/s'`` (default).

        normalize_gain : `bool`, optional
            Whether to normalize the gain when converting from Hz to rad/s.

            - `False` (default):
              Multiply zeros/poles by -2π but leave gain unchanged.
              This matches the LIGO GDS **'f' plane** convention
              (``plane='f'`` in ``s2z()``).

            - `True`:
              Normalize gain to preserve frequency response magnitude.
              Gain is scaled by :math:`|∏p_i/∏z_i| · (2π)^{(n_p - n_z)}`.
              Use this when your filter was designed with the transfer
              function :math:`H(f) = k·∏(f-z_i)/∏(f-p_i)` in Hz.
              This matches the LIGO GDS **'n' plane** convention
              (``plane='n'`` in ``s2z()``).

            Only used for analogue filters in Hz (``analog=True, unit="Hz"``).

        kwargs
            All other keyword arguments are passed to
            :meth:`~matplotlib.axes.Axes.plot`.

        Returns
        -------
        mag, phase : `matplotlib.lines.Line2D`
            The lines drawn for the magnitude and phase of the filter.

        Raises
        ------
        ValueError
            If ``analog=False`` is given and ``sample_rate`` isn't.
        """
        frequencies, fresp = filter_design.frequency_response(
            filter_,
            frequencies,
            analog=analog,
            sample_rate=sample_rate,
            unit=unit or "rad/s",
            normalize_gain=normalize_gain,
        )
        if frequencies[0] == 0:
            frequencies = frequencies[1:]
            fresp = fresp[1:]

        mag = abs(fresp)
        if dB:
            mag = _to_db(mag, power=False)
        if analog:
            phase = numpy.unwrap(numpy.arctan2(fresp.imag, fresp.real)) * 180.0 / pi
        else:
            phase = numpy.rad2deg(numpy.unwrap(numpy.angle(fresp)))

        # draw
        mline = self.maxes.plot(frequencies, mag, **kwargs)[0]
        pline = self.paxes.plot(frequencies, phase, **kwargs)[0]
        return mline, pline

    def add_frequencyseries(
        self,
        spectrum: FrequencySeries,
        *,
        dB: bool = True,  # noqa: N803
        power: bool = False,
        **kwargs,
    ) -> tuple[Line2D, Line2D]:
        """Plot the magnitude and phase of a complex-valued `FrequencySeries`.

        Parameters
        ----------
        spectrum : `~gwpy.frequencyseries.FrequencySeries`
            the (complex-valued) `FrequencySeries` to display

        dB : `bool`, optional
            If `True`, display magnitude in decibels, otherwise display
            amplitude.

        power : `bool`, optional
            Give `True` to incidate that ``spectrum`` holds power values,
            so ``dB = 10 * log(abs(spectrum))``, otherwise
            ``db = 20 * log(abs(spectrum))``.
            This argument is ignored if ``db=False``.

        **kwargs
            All other keyword arguments are passed to
            :meth:`~matplotlib.axes.Axes.plot`.

        Returns
        -------
        mag, phase : `matplotlib.lines.Line2D`
            The lines drawn for the magnitude and phase of the filter.
        """
        # parse spectrum arguments
        kwargs.setdefault("label", spectrum.name)

        # get magnitude
        mag = numpy.absolute(cast("Array1D", spectrum.value))
        if dB:
            mag = _to_db(mag, power=power)

        # get phase
        phase = numpy.angle(spectrum.value, deg=True)

        # plot
        w = spectrum.frequencies.value
        mline = self.maxes.plot(w, mag, **kwargs)[0]
        pline = self.paxes.plot(w, phase, **kwargs)[0]
        return mline, pline



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.cli.qtransform`."""

from __future__ import annotations

import os
import re
from typing import TYPE_CHECKING

import pytest

from .. import QtransformProduct
from .base import update_namespace
from .test_spectrogram import TestSpectrogramProduct

if TYPE_CHECKING:
    from pathlib import Path
    from typing import ClassVar


class TestQtransformProduct(TestSpectrogramProduct):
    """Tests for `gwpy.cli.QtransformProduct`."""

    TEST_CLASS = QtransformProduct
    ACTION = "qtransform"
    TEST_ARGS: ClassVar[list[str]] = [
        "--chan", "X1:TEST-CHANNEL",
        "--gps", "5",
        "--search", "10",
        "--nds2-server", "nds.test.gwpy",
        "--outdir", os.path.curdir,
        "--average-method", "median",
    ]

    def test_finalize_arguments(self, prod):
        """Test finalising arguments for `Qtransform`."""
        assert prod.start_list == [prod.args.gps - prod.args.search/2.]
        assert prod.duration == prod.args.search
        assert prod.args.color_scale == "linear"
        assert prod.args.xmin is None
        assert prod.args.xmax is None

    def test_init(self, args):
        """Test initialising a `Qtransform`."""
        update_namespace(args, qrange=(100., 110.))
        prod = self.TEST_CLASS(args)
        assert prod.qxfrm_args["gps"] == prod.args.gps
        assert prod.qxfrm_args["qrange"] == (100., 110.)

    def test_get_title(self, dataprod: QtransformProduct):
        """Test `QtransformProduct.get_title()`."""
        _float_reg = r"[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?"
        title_reg = re.compile(
            r"\A"
            fr"q: {_float_reg}, "
            fr"tres: {_float_reg}, "
            r"whitened, "
            fr"f-range: \[{_float_reg}, {_float_reg}\], "
            fr"e-range: \[{_float_reg}, {_float_reg}\]"
            r"\Z",
            re.IGNORECASE,
        )
        assert title_reg.match(dataprod.get_title())

    def test_get_suptitle(self, prod: QtransformProduct):
        """Test `QtransformProduct.get_suptitle()`."""
        assert prod.get_suptitle() == f"Q-transform: {prod.chan_list[0]}"

    @pytest.mark.requires("nds2")
    @pytest.mark.usefixtures("nds2_connection")
    def test_run(self, prod: QtransformProduct, tmp_path: Path):
        """Test `QtransformProduct.run()`."""
        prod.args.out = tmp_path
        prod.run()
        assert (tmp_path / "X1-TEST_CHANNEL-5.0-0.5.png").is_file()
        assert prod.plot_num == 1
        assert not prod.has_more_plots()



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.cli.coherence`."""

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Generic,
    TypeVar,
)

from ... import cli
from .base import _TestCliProduct
from .test_spectrum import TestSpectrumProduct as _TestSpectrumProduct

if TYPE_CHECKING:
    from typing import ClassVar

CoherenceProductType = TypeVar("CoherenceProductType", bound=cli.CoherenceProduct)


class TestCoherenceProduct(
    _TestSpectrumProduct[CoherenceProductType],
    Generic[CoherenceProductType],
):
    """Tests for `gwpy.cli.CoherenceProduct`."""

    TEST_CLASS: ClassVar[type[cli.CoherenceProduct]] = cli.CoherenceProduct
    ACTION = "coherence"
    TEST_ARGS: ClassVar[list[str]] = [
        *_TestCliProduct.TEST_ARGS,
        "--chan",
         "Y1:TEST-CHANNEL",
         "--secpfft", "0.25",
    ]

    def test_init_ref_chan(self, prod: CoherenceProductType):
        """Test that `CoherenceProduct.ref_chan` gets initialised properly."""
        assert prod.ref_chan == prod.chan_list[0]

    def test_get_suptitle(self, prod: CoherenceProductType):
        """Test `CoherenceProduct.get_suptitle`."""
        assert prod.get_suptitle() == f"Coherence: {prod.chan_list[0]}"



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for Table I/O."""

from __future__ import annotations

from functools import wraps
from typing import (
    TYPE_CHECKING,
    cast,
)

from ..filter import filter_table

if TYPE_CHECKING:
    from collections.abc import (
        Callable,
        Collection,
        Iterable,
        Mapping,
    )
    from typing import (
        ParamSpec,
        TypeVar,
    )

    import h5py
    import numpy
    from astropy.table import Table

    from ..filter import FilterLike

    P = ParamSpec("P")
    T = TypeVar("T", bound=Table)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


# -- dynamic column helpers ----------

def dynamic_columns(
    columns: Iterable[str] | None,
    valid_columns: Collection[str],
    dynamic_column_map: Mapping[str, Iterable[str]],
) -> tuple[set[str] | None, set[str]]:
    """Return the list of columns to read and those to dynamically create.

    Parameters
    ----------
    columns : `list` of `str` or `None`
        The columns requested for the final table.
        If `None` is given, `None` is returned.

    valid_columns : `list` of `str`
        The list of columns that are valid for the table when in the
        relevant format.

    dynamic_column_map: `dict` of `str`, `set` pairs
        The mapping of dynamic column names to columns that need to be read
        to support that column.

    Returns
    -------
    read_columns : `set` of `str`, `None`
        The set of columns to read. Will be `None` if ``columns`` is `None`.

    dynamic_columns : `set` of `str`
        The set of columns to create dynamically after reading.

    Notes
    -----
    The set of column names that is the `~set.difference` between
    the returned ``read_columns`` and ``dynamic_columns`` should be removed
    from the table after the dynamic columns are generated.
    """
    if columns is None:
        return None, set()
    read = set()
    dynamic = set()
    for name in iter(columns):
        # column name is present in the file
        if name in valid_columns:
            read.add(name)
        # column name is a dynamic column that is generated after reading
        elif name in dynamic_column_map:
            dynamic.add(name)
            read.update(dynamic_column_map[name])
        else:
            names = list(valid_columns or []) + list(dynamic_column_map)
            msg = (
                f"'{name}' is not a valid column name; "
                f"valid column names: {', '.join(names)}"
            )
            raise ValueError(msg)
    return read, dynamic


def mtotal(
    table: h5py.Group | Table,
) -> numpy.ndarray:
    """Calculate the total mass column for this table."""
    mass1 = table["mass1"][:]
    mass2 = table["mass2"][:]
    return mass1 + mass2


def mchirp(
    table: h5py.Group | Table,
) -> numpy.ndarray:
    """Calculate the chirp mass column for this table."""
    mass1 = table["mass1"][:]
    mass2 = table["mass2"][:]
    return (mass1 * mass2) ** (3/5.) / (mass1 + mass2) ** (1/5.)


DYNAMIC_COLUMN_FUNC: dict[str, Callable] = {
    "mchirp": mchirp,
    "mtotal": mtotal,
}
DYNAMIC_COLUMN_INPUT: dict[str, set[str]] = {
    "mchirp": {"mass1", "mass2"},
    "mtotal": {"mass1", "mass2"},
}


# -- table i/o utilities -------------

def read_with_columns(
    func: Callable[P, T],
) -> Callable[P, T]:
    """Decorate a Table read method to use the ``columns`` keyword."""
    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        # parse columns argument
        columns = kwargs.pop("columns", None)

        # read table
        tab = func(*args, **kwargs)

        # filter on columns
        if columns is None:
            return tab
        try:
            return tab[columns]
        except KeyError as exc:
            missing = str(exc)
            names = list(tab.colnames)
            msg = (
                f"column {missing} not found; "
                f"valid column names: {', '.join(names)}"
            )
            raise ValueError(msg) from exc


    return wrapper


def read_with_where(
    func: Callable[P, T],
) -> Callable[P, T]:
    """Decorate a Table read method to apply the ``where`` keyword.

    Allows for filtering tables on-the-fly when read.
    """
    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        """Execute a function, then apply a where filter."""
        # parse where
        where = cast(
            "FilterLike | Iterable[FilterLike]",
            kwargs.pop("where", None) or [],
        )

        # read table
        tab = func(*args, **kwargs)

        # apply where
        if where:
            return filter_table(tab, where)

        return tab

    return wrapper


def read_with_columns_and_where(
    func: Callable[P, T],
) -> Callable[P, T]:
    """Decorate a read function to support both ``columns`` and ``where``.

    The ``where`` decorator is applied _first_ so that the conditions can
    be applied to columns that _aren't_ requested.
    """
    return read_with_columns(read_with_where(func))
